import random
import typing as t

from aladindb.backends.base.query import Select
from aladindb.misc.special_dicts import MongoStyleDict


class ExpiryCache(list):
    def __getitem__(self, index):
        item = super().__getitem__(index)
        del self[index]
        return item


class QueryDataset:
    """
    A dataset class which can be used to define a torch dataset class.

    :param select: A select query object which defines the query to be executed.
    :param keys: A list of keys to be returned from the dataset.
    :param fold: The fold to be used for the dataset.
    :param suppress: A list of keys to be suppressed from the dataset.
    :param transform: A callable which can be used to transform the dataset.
    :param db: A ``DB`` object to be used for the dataset.
    :param ids: A list of ids to be used for the dataset.
    :param in_memory: A boolean flag to indicate if the dataset should be loaded
                      in memory.
    :param extract: A key to be extracted from the dataset.
    """

    def __init__(
        self,
        select: Select,
        keys: t.Optional[t.List[str]] = None,
        fold: t.Union[str, None] = 'train',
        suppress: t.Sequence[str] = (),
        transform: t.Optional[t.Callable] = None,
        db=None,
        ids: t.Optional[t.List[str]] = None,
        in_memory: bool = True,
        extract: t.Optional[str] = None,
        **kwargs,
    ):
        self._database = db
        self.keys = keys

        self.transform = transform if transform else lambda x: x
        if fold is not None:
            self.select = select.add_fold(fold)
        else:
            self.select = select
        self.in_memory = in_memory
        if self.in_memory:
            if ids is None:
                self._documents = list(self.database.execute(self.select))
            else:
                self._documents = list(
                    self.database.execute(self.select.select_using_ids(ids))
                )
        else:
            if ids is None:
                self._ids = [
                    r[self.select.id_field]
                    for r in self.database.execute(self.select.select_ids)
                ]
            else:
                self._ids = ids
            self.select_one = self.select.select_single_id
        self.suppress = suppress
        self.extract = extract

    @property
    def database(self):
        if self._database is None:
            from aladindb.base.build import build_datalayer

            self._database = build_datalayer()
        return self._database

    def __len__(self):
        if self.in_memory:
            return len(self._documents)
        else:
            return len(self._ids)

    def __getitem__(self, item):
        if self.in_memory:
            input = self._documents[item]
        else:
            input = self.select_one(
                self._ids[item], self.database, encoders=self.database.encoders
            )
        r = MongoStyleDict(input.unpack())
        s = MongoStyleDict({})

        if self.keys is not None:
            for k in self.keys:
                if k == '_base':
                    s[k] = r
                else:
                    s[k] = r[k]
        else:
            s = r
        out = self.transform(s)
        if self.extract:
            out = out[self.extract]
        return out


class CachedQueryDataset:
    """
    This class which fetch the document corresponding to the given ``index``.
    This class prefetches documents from database and stores in the memory.

    This can drastically reduce database read operations and hence reduce the overall
    load on the database.
    """

    _BACKFILL_INDEX = 0.2

    def __init__(
        self,
        select: Select,
        keys=None,
        fold='train',
        suppress=(),
        transform=None,
        database=None,
        prefetch_size: int = 100,
    ):
        self._database = database
        self.keys = keys

        self.transform = transform if transform else lambda x: x
        self.select = select.add_fold(fold)

        self.ids = [doc.id for doc in self.database.execute(self.select.select_ids)]
        self.suppress = suppress
        self._max_cache_size = prefetch_size
        self._cache: ExpiryCache = self._fetch_cache()
        self._total_documents = self.count_documents()

    def count_documents(self) -> int:
        """Return the number of matching documents"""
        return self.database.execute(self.select).count()

    def _fetch_cache(self):
        cache_ids = random.sample(self.ids, self._max_cache_size)
        return ExpiryCache(
            list(self.database.execute(self.select.select_using_ids(cache_ids)))
        )

    @property
    def database(self):
        if self._database is None:
            from aladindb.base.build import build_datalayer

            self._database = build_datalayer()
        return self._database

    def _unpack(self, documents):
        batch = []
        for document in documents:
            r = MongoStyleDict(document.unpack())
            s = MongoStyleDict({})

            if self.keys is not None:
                for k in self.keys:
                    if k == '_base':
                        s[k] = r
                    else:
                        s[k] = r[k]
            else:
                s = r
            s = self.transform(s)
            batch.append(s)
        return batch

    def _get_random_index(self, index):
        return int((index * len(self._cache)) / self._total_documents)

    def _backfill_cache(self):
        if len(self._cache) <= int(self._max_cache_size * self._BACKFILL_INDEX):
            self._cache = self._fetch_cache()

    def __len__(self):
        return self._total_documents

    def __getitem__(self, index):
        index = self._get_random_index(index)
        document = self._cache[index]
        self._backfill_cache()
        r = MongoStyleDict(document.unpack())
        s = MongoStyleDict({})

        if self.keys is not None:
            for k in self.keys:
                if k == '_base':
                    s[k] = r
                else:
                    s[k] = r[k]
        else:
            s = r
        return self.transform(s)


def query_dataset_factory(data_prefetch: bool = False, **kwargs):
    if data_prefetch:
        return CachedQueryDataset(**kwargs)
    return QueryDataset(**kwargs)



import typing as t
from abc import ABC, abstractmethod

from aladindb.components.model import APIModel, Model


class BaseDataBackend(ABC):
    def __init__(self, conn: t.Any, name: str):
        self.conn = conn
        self.name = name
        self.in_memory: bool = False

    @property
    def db(self):
        raise NotImplementedError

    @abstractmethod
    def url(self):
        """
        Databackend connection url
        """
        pass

    @abstractmethod
    def build_metadata(self):
        """
        Build a default metadata store based on current connection.
        """
        pass

    @abstractmethod
    def build_artifact_store(self):
        """
        Build a default artifact store based on current connection.
        """
        pass

    def create_model_table_or_collection(self, model: t.Union[Model, APIModel]):
        pass

    @abstractmethod
    def get_table_or_collection(self, identifier):
        pass

    def set_content_bytes(self, r, key, bytes_):
        raise NotImplementedError

    @abstractmethod
    def drop(self, force: bool = False):
        """
        Drop the databackend.
        """
        pass

    @abstractmethod
    def disconnect(self):
        """
        Disconnect the client
        """



import dataclasses as dc
import typing as t
from functools import cached_property
from queue import Empty, Queue

from .runnable import Runnable
from .thread import HasThread, none

_SENTINEL_MESSAGE = object()


class HasRunnables(Runnable):
    """Collect zero or more Runnable into one"""

    runnables: t.Sequence[Runnable]

    def start(self):
        for r in self.runnables:
            r.running.on_set.append(self._on_start)
            r.stopped.on_set.append(self._on_stop)
            r.start()

    def stop(self):
        self.running.clear()
        for r in self.runnables:
            r.stop()

    def finish(self):
        for r in self.runnables:
            r.finish()

    def join(self, timeout: t.Optional[float] = None):
        for r in self.runnables:
            r.join(timeout)

    def _on_start(self):
        if not self.running and all(r.running for r in self.runnables):
            super().start()

    def _on_stop(self):
        if not self.stopped and all(r.stopped for r in self.runnables):
            super().stop()


@dc.dataclass
class ThreadQueue(HasRunnables):
    """A simple multi-producer, multi-consumer queue with one thread per consumer.

    There is a special `finish_message` value, which when received shuts down
    that consumer.  ThreadQueue.finish() puts one `self.finish_message` onto the
    queue for each consumer.
    """

    callback: t.Callable[[t.Any], None]
    error: t.Callable = none
    maxsize: int = 0
    name: str = 'thread_queue'
    thread_count: int = 1
    timeout: t.Optional[float] = 0.1

    def __post_init__(self):
        HasRunnables.__init__(self)
        self.runnables = tuple(self._thread(i) for i in range(self.thread_count))

    @cached_property
    def queue(self) -> Queue:
        return Queue(self.maxsize)

    def finish(self) -> None:
        """Put an empty message into the queue for each listener"""
        for _ in self.runnables:
            self.queue.put(_SENTINEL_MESSAGE)

    def _thread(self, i: int) -> HasThread:
        thread = HasThread(name=f'{self.name}-{i}', error=self.error)

        def callback():
            self.running.wait()
            while self.running and thread.running:
                try:
                    item = self.queue.get(timeout=self.timeout)
                except Empty:
                    continue
                if item is _SENTINEL_MESSAGE:
                    return
                try:
                    self.callback(item)
                except Exception:
                    self.stop()
                    raise

        thread.callback = callback
        return thread



import time
import typing as t
from abc import ABC, abstractmethod

if t.TYPE_CHECKING:
    from aladindb.backends.base.query import Select


class NonExistentMetadataError(Exception):
    ...


class MetaDataStore(ABC):
    """
    Abstraction for storing meta-data separately from primary data.

    :param conn: connection to the meta-data store
    :param name: Name to identify DB using the connection
    """

    def __init__(
        self,
        conn: t.Any,
        name: t.Optional[str] = None,
    ):
        self.name = name
        self.conn = conn

    @abstractmethod
    def url(self):
        """
        Metadata store connection url
        """
        pass

    @abstractmethod
    def create_component(self, info: t.Dict):
        """
        Create a component in the metadata store.

        :param info: dictionary containing information about the component.
        """
        pass

    @abstractmethod
    def create_job(self, info: t.Dict):
        """
        Create a job in the metadata store.
        :param info: dictionary containing information about the job.
        """
        pass

    @abstractmethod
    def create_parent_child(self, parent: str, child: str):
        """
        Create a parent-child relationship between two components.

        :param parent: parent component
        :param child: child component
        """
        pass

    @abstractmethod
    def drop(self, force: bool = False):
        """
        Drop the metadata store.

        :param force: whether to force the drop (without confirmation)
        """
        pass

    # ------------------ COMPONENTS ------------------

    @abstractmethod
    def get_job(self, job_id: str):
        """
        Get a job from the metadata store.

        :param job_id: job identifier
        """
        pass

    @abstractmethod
    def update_job(self, job_id: str, key: str, value: t.Any):
        """
        Update a job in the metadata store.

        :param job_id: job identifier
        :param key: key to be updated
        :param value: value to be updated
        """
        pass

    def watch_job(self, identifier: str):
        """
        Listen to a job.

        :param identifier: job identifier
        """
        status = 'pending'
        n_lines = 0
        n_lines_stderr = 0
        while status in {'pending', 'running'}:
            r = self.get_job(identifier)
            status = r['status']
            if status == 'running':
                if len(r['stdout']) > n_lines:
                    print(''.join(r['stdout'][n_lines:]), end='')
                    n_lines = len(r['stdout'])
                if len(r['stderr']) > n_lines_stderr:
                    print(''.join(r['stderr'][n_lines_stderr:]), end='')
                    n_lines_stderr = len(r['stderr'])
            time.sleep(0.01)
        r = self.get_job(identifier)
        if status == 'success':
            if len(r['stdout']) > n_lines:
                print(''.join(r['stdout'][n_lines:]), end='')
            if len(r['stderr']) > n_lines_stderr:
                print(''.join(r['stderr'][n_lines_stderr:]), end='')
        elif status == 'failed':
            print(r['msg'])

    @abstractmethod
    def show_jobs(self):
        """
        Show all jobs in the metadata store.
        """
        pass

    @abstractmethod
    def show_components(self, type_id: str, **kwargs):
        """
        Show all components in the metadata store.

        :param type_id: type of component
        :param **kwargs: additional arguments
        """
        pass

    @abstractmethod
    def show_component_versions(self, type_id: str, identifier: str):
        """
        Show all versions of a component in the metadata store.

        :param type_id: type of component
        :param identifier: identifier of component
        """
        pass

    def get_indexing_listener_of_vector_index(
        self, identifier: str, version: t.Optional[int] = None
    ):
        info = self.get_component(
            'vector_index', identifier=identifier, version=version
        )
        indexing_listener = info['dict']['indexing_listener']['identifier']
        return indexing_listener

    @abstractmethod
    def delete_component_version(self, type_id: str, identifier: str, version: int):
        """
        Delete a component version from the metadata store.

        :param type_id: type of component
        :param identifier: identifier of component
        :param version: version of component
        """
        pass

    @abstractmethod
    def component_version_has_parents(
        self, type_id: str, identifier: str, version: int
    ):
        """
        Check if a component version has parents.

        :param type_id: type of component
        :param identifier: identifier of component
        :param version: version of component
        """
        pass

    @abstractmethod
    def get_metadata(self, key):
        """
        Get metadata from the metadata store.

        :param key: key of metadata
        """
        pass

    @abstractmethod
    def get_latest_version(
        self, type_id: str, identifier: str, allow_hidden: bool = False
    ):
        """
        Get the latest version of a component.

        :param type_id: type of component
        :param identifier: identifier of component
        :param allow_hidden: whether to allow hidden components
        """
        pass

    @abstractmethod
    def _get_component(
        self,
        type_id: str,
        identifier: str,
        version: int,
        allow_hidden: bool = False,
    ):
        pass

    def get_component(
        self,
        type_id: str,
        identifier: str,
        version: t.Optional[int] = None,
        allow_hidden: bool = False,
    ) -> t.Dict[str, t.Any]:
        """
        Get a component from the metadata store.

        :param type_id: type of component
        :param identifier: identifier of component
        :param version: version of component
        :param allow_hidden: whether to allow hidden components
        """
        if version is None:
            version = self.get_latest_version(
                type_id=type_id, identifier=identifier, allow_hidden=allow_hidden
            )
        r = self._get_component(
            type_id=type_id,
            identifier=identifier,
            version=version,
            allow_hidden=allow_hidden,
        )
        if r is None:
            raise FileNotFoundError(f'Object {identifier} does not exist in metadata')
        return r

    @abstractmethod
    def _update_object(
        self,
        identifier: str,
        type_id: str,
        key: str,
        value: t.Any,
        version: int,
    ):
        pass

    def update_object(
        self,
        identifier: str,
        type_id: str,
        key: str,
        value: str,
        version: t.Optional[int] = None,
    ):
        """
        Update an object in the metadata store.

        :param identifier: identifier of object
        :param type_id: type of object
        :param key: key to be updated
        :param value: value to be updated
        :param version: version of object
        """
        if version is not None:
            version = self.get_latest_version(type_id, identifier)
        assert isinstance(version, int)
        return self._update_object(
            identifier=identifier,
            type_id=type_id,
            key=key,
            value=value,
            version=version,
        )

    @abstractmethod
    def _replace_object(self, info, identifier, type_id, version):
        pass

    def replace_object(
        self,
        info: t.Dict[str, t.Any],
        identifier: str,
        type_id: str,
        version: t.Optional[int] = None,
    ) -> None:
        """
        Replace an object in the metadata store.

        :param info: dictionary containing information about the object
        :param identifier: identifier of object
        :param type_id: type of object
        :param version: version of object
        """
        if version is not None:
            version = self.get_latest_version(type_id, identifier)
        return self._replace_object(
            info=info,
            identifier=identifier,
            type_id=type_id,
            version=version,
        )

    @abstractmethod
    def write_output_to_job(self, identifier: str, msg: str, stream: str):
        """
        Write output to a job in the metadata store.

        :param identifier: identifier of job
        :param msg: message to be written
        :param stream: stream to be written to
        """
        pass

    @abstractmethod
    def get_component_version_parents(self, unique_id: str):
        """
        Get the parents of a component version.

        :param unique_id: unique identifier of component version
        """
        pass

    @abstractmethod
    def hide_component_version(self, type_id: str, identifier: str, version: int):
        """
        Hide a component version.

        :param type_id: type of component
        :param identifier: identifier of component
        :param version: version of component
        """
        pass

    @abstractmethod
    def update_metadata(self, key, value):
        """
        Update metadata in the metadata store.

        :param key: Key of metadata
        :param value: Value of metadata
        """
        pass

    def add_query(self, query: 'Select', model: str):
        """
        Add query id to query table
        """
        raise NotImplementedError

    def get_queries(self, model: str):
        """
        Get all queries from query table corresponding
        to the model.
        """

    @abstractmethod
    def disconnect(self):
        """
        Disconnect the client
        """



import functools
import sys
import typing as t

import pydantic

__all__ = 'Factory', 'JSONable'

TYPE_ID_ATTR = 'type_id'
_NONE = object()


def Factory(factory: t.Callable, **ka) -> t.Any:
    return pydantic.Field(default_factory=factory, **ka)


class JSONable(pydantic.BaseModel):
    """
    A base class for classes that can be
    converted to and from JSON
    """

    class Config:
        # Fail in deserializion if there are extra fields
        extra = 'forbid'

        # See https://github.com/samuelcolvin/pydantic/issues/1241
        if pydantic.__version__.startswith('2'):
            ignored_types = (functools.cached_property,)
        else:
            keep_untouched = (functools.cached_property,)

    SUBCLASSES: t.ClassVar[t.Set[t.Type]] = set()
    TYPE_ID_TO_CLASS: t.ClassVar[t.Dict[str, t.Type]] = {}

    @functools.wraps(pydantic.BaseModel.dict)
    def dict(self, *a, **ka):
        d = super().dict(*a, **ka)
        properties = self.schema()['properties']
        return {k: v for k, v in d.items() if k in properties}

    def deepcopy(self) -> 'JSONable':
        return self.copy(deep=True)

    def __init_subclass__(cls, *a, **ka):
        super().__init_subclass__(*a, **ka)
        cls.SUBCLASSES.add(cls)

        try:
            type_id = getattr(cls, TYPE_ID_ATTR)
        except AttributeError:
            return

        if not isinstance(type_id, str):
            (type_id,) = type_id.__args__

        if old_model := cls.TYPE_ID_TO_CLASS.get(type_id):
            print(f'Duplicate type_id: old={old_model}, new={cls}', file=sys.stderr)

        cls.TYPE_ID_TO_CLASS[type_id] = cls



import json
import sys
import threading
import time
from functools import cached_property
from traceback import format_exc

import uvicorn
from fastapi import APIRouter, Depends, FastAPI, Request
from fastapi.responses import JSONResponse
from prettytable import PrettyTable
from starlette.middleware.base import BaseHTTPMiddleware

from aladindb import logging
from aladindb.base.build import build_datalayer
from aladindb.base.datalayer import Datalayer

# --------------- Create exception handler middleware-----------------


class ExceptionHandlerMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        try:
            return await call_next(request)
        except Exception as e:
            host = getattr(getattr(request, "client", None), "host", None)
            port = getattr(getattr(request, "client", None), "port", None)
            url = (
                f"{request.url.path}?{request.query_params}"
                if request.query_params
                else request.url.path
            )
            exception_type, exception_value, _ = sys.exc_info()
            exception_traceback = format_exc()
            exception_name = getattr(exception_type, "__name__", None)
            msg = f'{host}:{port} - "{request.method} {url}"\
                    500 Internal Server Error <{exception_name}:\
                    {exception_value}>'
            logging.exception(msg, e=e)
            return JSONResponse(
                status_code=500,
                content={
                    'error': exception_name,
                    'messages': msg,
                    'traceback': exception_traceback,
                },
            )


class aladinApp:
    """
    This is a wrapper class that prepares helper functions used to create a
    fastapi application in the realm of aladindb.
    """

    def __init__(self, service='vector_search', port=8000):
        self.service = service
        self.port = port

        self.app_host = '0.0.0.0'
        self._app = FastAPI()

        self.router = APIRouter()
        self._user_startup = False
        self._user_shutdown = False

        self._app.add_middleware(ExceptionHandlerMiddleware)

    @cached_property
    def app(self):
        self._app.include_router(self.router)
        return self._app

    @cached_property
    def db(self):
        return self._app.state.pool

    def add(self, *args, method='post', **kwargs):
        """
        Register an endpoint with this method.
        """

        def decorator(function):
            self.router.add_api_route(
                *args, **kwargs, endpoint=function, methods=[method]
            )

        return decorator

    def add_default_endpoints(self):
        """
        Add a list of default endpoints which comes out of the box with `aladinApp`
        """

        @self.router.get('/health')
        def health():
            return {'status': 200}

        @self.router.post('/handshake/config')
        def handshake(cfg: str):
            from aladindb import CFG

            cfg_dict = json.loads(cfg)
            if CFG.match(cfg_dict):
                return {'status': 200, 'msg': 'matched'}

            return JSONResponse(
                status_code=400,
                content={'error': 'Config is not match'},
            )

    def print_routes(self):
        table = PrettyTable()

        # Define the table headers
        table.field_names = ["Path", "Methods", "Function"]

        # Add rows to the table
        for route in self._app.routes:
            table.add_row([route.path, ", ".join(route.methods), route.name])

        logging.info(f"Routes for '{self.service}' app: \n{table}")

    def pre_start(self, cfg=None):
        self.add_default_endpoints()

        if not self._user_startup:
            self.startup(cfg=cfg)
        if not self._user_shutdown:
            self.shutdown()
        assert self.app

    def start(self):
        """
        This method is used to start the application server
        """
        self.pre_start()

        self.print_routes()

        uvicorn.run(
            self._app,
            host=self.app_host,
            port=self.port,
            reload=False,
        )

    def startup(self, function=None, cfg=None):
        """
        This method is used to register a startup function
        """
        self._user_startup = True

        @self._app.on_event('startup')
        def startup_db_client():
            sys.path.append('./')
            db = build_datalayer(cfg)
            db.server_mode = True
            if function:
                function(db=db)
            self._app.state.pool = db

        return

    def shutdown(self, function=None):
        """
        This method is used to register a shutdown function
        """
        self._user_shutdown = True

        @self._app.on_event('shutdown')
        def shutdown_db_client():
            try:
                if function:
                    function(db=self._app.state.pool)
                self._app.state.pool.close()
            except AttributeError:
                raise Exception('Could not close the database properly')


def database(request: Request) -> Datalayer:
    return request.app.state.pool


def DatalayerDependency():
    """
    A helper method to be used for injecting datalayer instance
    into endpoint implementation
    """
    return Depends(database)


class Server(uvicorn.Server):
    def install_signal_handlers(self):
        pass

    def run_in_thread(self):
        self._thread = threading.Thread(target=self.run)
        self._thread.start()

        while not self.started:
            time.sleep(1e-3)

    def stop(self):
        self.should_exit = True
        self._thread.join()



import io

import PIL.Image
import PIL.JpegImagePlugin
import PIL.PngImagePlugin

from aladindb.components.encoder import Encoder

BLANK_IMAGE = PIL.Image.new('RGB', (600, 600), (255, 255, 255))


def encode_pil_image(x):
    buffer = io.BytesIO()
    x.save(buffer, 'png')
    return buffer.getvalue()


class DecoderPILImage:
    def __init__(self, handle_exceptions: bool = True):
        self.handle_exceptions = handle_exceptions

    def __call__(self, bytes):
        try:
            return PIL.Image.open(io.BytesIO(bytes))
        except Exception as e:
            if self.handle_exceptions:
                return BLANK_IMAGE
            else:
                raise e


decode_pil_image = DecoderPILImage()


pil_image = Encoder(
    'pil_image',
    encoder=encode_pil_image,
    decoder=decode_pil_image,
)



import dataclasses as dc
import typing as t

from aladindb import CFG
from aladindb.base.document import Document
from aladindb.components.encoder import Encoder
from aladindb.misc.files import load_uris


@dc.dataclass
class aladinCursor:
    """
    A cursor that wraps a cursor and returns ``Document`` wrapping
    a dict including ``Encodable`` objects.

    :param raw_cursor: the cursor to wrap
    :param id_field: the field to use as the document id
    :param encoders: a dict of encoders to use to decode the documents
    :param scores: a dict of scores to add to the documents
    """

    raw_cursor: t.Any
    id_field: str
    encoders: t.Dict[str, Encoder] = dc.field(default_factory=dict)
    scores: t.Optional[t.Dict[str, float]] = None

    _it: int = 0

    def limit(self, *args, **kwargs) -> 'aladinCursor':
        """
        Limit the number of results returned by the cursor.
        """
        return aladinCursor(
            raw_cursor=self.raw_cursor.limit(*args, **kwargs),
            id_field=self.id_field,
            encoders=self.encoders,
            scores=self.scores,
        )

    def cursor_next(self):
        if isinstance(self.raw_cursor, list):
            if self._it >= len(self.raw_cursor):
                raise StopIteration
            r = self.raw_cursor[self._it]
            self._it += 1
            return r
        return self.raw_cursor.next()

    def wrap_document(self, r, encoders):
        """
        Wrap a document in a ``Document``.
        """
        return Document(Document.decode(r, encoders))

    def __iter__(self):
        return self

    def __next__(self):
        r = self.cursor_next()
        if self.scores is not None:
            r['score'] = self.scores[str(r[self.id_field])]
        if CFG.hybrid_storage:
            load_uris(r, CFG.downloads_folder, encoders=self.encoders)
        return self.wrap_document(r, self.encoders)

    next = __next__



import dataclasses as dc
import functools
import random
import re
import typing as t
import warnings

from transformers import (
    DataCollatorWithPadding,
    Pipeline as BasePipeline,
    Trainer,
    TrainingArguments,
    pipeline as _pipeline,
)

from aladindb import logging
from aladindb.backends.base.query import Select
from aladindb.backends.query_dataset import query_dataset_factory
from aladindb.base.artifact import Artifact
from aladindb.base.datalayer import Datalayer
from aladindb.components.metric import Metric
from aladindb.components.model import Model, _TrainingConfiguration
from aladindb.misc.special_dicts import MongoStyleDict

_DEFAULT_PREFETCH_SIZE: int = 100


@functools.wraps(TrainingArguments)
def TransformersTrainerConfiguration(identifier: str, *args, **kwargs):
    if 'output_dir' in kwargs:
        cfg = TrainingArguments(**kwargs)
    else:
        cfg = TrainingArguments(output_dir=args[0], **kwargs)
    return _TrainingConfiguration(identifier=identifier, kwargs=cfg.to_dict())


@dc.dataclass
class Pipeline(Model):
    """A wrapper for ``transformers.Pipeline``

    :param preprocess_type: The type of preprocessing to use {'tokenizer'}
    :param preprocess_kwargs: The type of preprocessing to use. Currently only
    :param postprocess_kwargs: The type of postprocessing to use.
    :param task: The task to use for the pipeline.
    """

    preprocess_type: str = 'tokenizer'
    preprocess_kwargs: t.Dict[str, t.Any] = dc.field(default_factory=dict)
    postprocess_kwargs: t.Dict[str, t.Any] = dc.field(default_factory=dict)
    task: str = 'text-classification'

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.object.artifact, BasePipeline):
            assert self.preprocess is None
            if self.preprocess_type == 'tokenizer':
                self.preprocess = self.object.artifact.tokenizer
            else:
                raise NotImplementedError(
                    'Only tokenizer is supported for now in pipeline mode'
                )
            self.object = Artifact(artifact=self.object.artifact.model)
            self.task = self.object.artifact.task
        if (
            self.collate_fn is None
            and self.preprocess is not None
            and self.preprocess_type == 'tokenizer'
        ):
            self.collate_fn = Artifact(
                DataCollatorWithPadding(self.preprocess.artifact),
                hash=random.randrange(1000000),
            )
        if not self.device:
            import torch

            self.device = "cuda" if torch.cuda.is_available() else "cpu"

    @functools.cached_property
    def pipeline(self):
        if self.preprocess_type == 'tokenizer':
            return _pipeline(
                self.task,
                model=self.object.artifact,
                tokenizer=self.preprocess.artifact,
            )
        else:
            warnings.warn('Only tokenizer is supported for now in pipeline mode')

    def _predict_with_preprocess_object_post(self, X, **kwargs):
        X = self.preprocess.artifact(X, **self.preprocess_kwargs)
        X = getattr(self.object.artifact, self.predict_method)(**X, **kwargs)
        X = getattr(self, 'postprocess', Artifact(lambda x: x)).artifact(
            X, **self.postprocess_kwargs
        )
        return X

    @functools.cached_property
    def training_arguments(self):
        return TrainingArguments(**self.configuration.kwargs)

    def _get_data(
        self,
        db,
        valid_data=None,
        data_prefetch: bool = False,
        prefetch_size: int = 100,
        X: str = '',
    ):
        def transform_function(r):
            text = r[X]
            r.update(**self.preprocess.artifact(text, **self.preprocess_kwargs))
            return r

        train_data = query_dataset_factory(
            select=self.training_select,
            keys=self.training_keys,
            fold='train',
            transform=transform_function,
            data_prefetch=data_prefetch,
            prefetch_size=prefetch_size,
            db=db,
        )
        valid_data = query_dataset_factory(
            select=self.training_select,
            keys=self.training_keys,
            fold='valid',
            transform=transform_function,
            data_prefetch=data_prefetch,
            prefetch_size=prefetch_size,
            db=db,
        )

        return train_data, valid_data

    def _fit(  # type: ignore[override]
        self,
        X: str,
        y: str,
        configuration: t.Optional[_TrainingConfiguration] = None,
        data_prefetch: bool = False,
        db: t.Optional[Datalayer] = None,
        metrics: t.Optional[t.Sequence[Metric]] = None,
        prefetch_size: int = _DEFAULT_PREFETCH_SIZE,
        select: t.Optional[Select] = None,
        validation_sets: t.Optional[t.Sequence[str]] = None,
        **kwargs,
    ) -> None:
        if configuration is not None:
            self.configuration = configuration
        if select is not None:
            self.training_select = select
        if validation_sets is not None:
            self.validation_sets = validation_sets
        if metrics is not None:
            self.metrics = metrics
        self.train_X = X
        self.train_y = y

        if isinstance(X, str):
            train_data, valid_data = self._get_data(
                db,
                valid_data=validation_sets,
                data_prefetch=data_prefetch,
                prefetch_size=prefetch_size,
                X=X,
            )

        def compute_metrics(eval_pred):
            output = {}
            for vs in validation_sets:
                vs = db.load('dataset', vs)
                unpacked = [MongoStyleDict(r.unpack()) for r in vs.data]
                predictions = self._predict([r[X] for r in unpacked])
                targets = [r[y] for r in unpacked]
                for m in metrics:
                    output[f'{vs.identifier}/{m.identifier}'] = m(predictions, targets)
            self.append_metrics(output)
            return output

        assert isinstance(self.collate_fn, Artifact)
        assert db is not None
        assert isinstance(self.object, Artifact)
        trainer = _TrainerWithSaving(
            model=self.object.artifact,
            args=self.training_arguments,
            train_dataset=train_data,
            eval_dataset=valid_data,
            data_collator=self.collate_fn.artifact,
            custom_saver=lambda: db.replace(self, upsert=True),
            compute_metrics=compute_metrics,
            **kwargs,
        )
        trainer.train()

    def _predict(self, X, one: bool = False, **kwargs):
        if self.pipeline is not None:
            out = self.pipeline(X, **self.preprocess_kwargs, **kwargs)
            out = [r['label'] for r in out]
            for i, p in enumerate(out):
                if re.match(r'^LABEL_[0-9]+', p):
                    out[i] = int(p[6:])
        else:
            out = self._predict_with_preprocess_object_post(X, **kwargs)
        if one:
            return out[0]
        return out


class _TrainerWithSaving(Trainer):
    def __init__(self, custom_saver=None, **kwargs):
        super().__init__(**kwargs)
        self.custom_saver = custom_saver

    def _save_checkpoint(self, model, trial, metrics=None):
        super()._save_checkpoint(model, trial, metrics=metrics)
        logging.info('Saving model...')
        if self.custom_saver:
            self.custom_saver()



import dataclasses as dc
import typing as t

from ibis.expr.datatypes import DataType, dtype as _dtype

from aladindb.base.serializable import Serializable


@dc.dataclass
class FieldType(Serializable):
    identifier: t.Union[str, DataType]

    def __post_init__(self):
        if isinstance(self.identifier, DataType):
            self.identifier = self.identifier.name


def dtype(x):
    return FieldType(_dtype(x))



import dataclasses as dc
import math
import random
import typing as t
import warnings
from collections import defaultdict

import click
import networkx
import tqdm

import aladindb as s
from aladindb import logging
from aladindb.backends.base.artifact import ArtifactStore
from aladindb.backends.base.backends import vector_searcher_implementations
from aladindb.backends.base.compute import ComputeBackend
from aladindb.backends.base.data_backend import BaseDataBackend
from aladindb.backends.base.metadata import MetaDataStore
from aladindb.backends.base.query import Delete, Insert, RawQuery, Select, Update
from aladindb.backends.ibis.query import Table
from aladindb.backends.local.compute import LocalComputeBackend
from aladindb.base import exceptions, serializable
from aladindb.base.cursor import aladinCursor
from aladindb.base.document import Document
from aladindb.base.aladin import aladin
from aladindb.cdc.cdc import DatabaseChangeDataCapture
from aladindb.components.component import Component
from aladindb.components.encoder import Encodable, Encoder
from aladindb.components.model import Model
from aladindb.components.serializer import serializers
from aladindb.jobs.job import ComponentJob, FunctionJob, Job
from aladindb.jobs.task_workflow import TaskWorkflow
from aladindb.misc.colors import Colors
from aladindb.misc.data import ibatch
from aladindb.misc.download import Downloader, download_content, gather_uris
from aladindb.misc.special_dicts import MongoStyleDict
from aladindb.vector_search.base import BaseVectorSearcher, VectorItem
from aladindb.vector_search.interface import FastVectorSearcher
from aladindb.vector_search.update_tasks import copy_vectors, delete_vectors

DBResult = t.Any
TaskGraph = t.Any

DeleteResult = DBResult
InsertResult = t.Tuple[DBResult, t.Optional[TaskGraph]]
SelectResult = aladinCursor
UpdateResult = t.Any

ExecuteQuery = t.Union[Select, Delete, Update, Insert, str]
ExecuteResult = t.Union[SelectResult, DeleteResult, UpdateResult, InsertResult]

ENDPOINTS = 'delete', 'execute', 'insert', 'like', 'select', 'update'


class Datalayer:
    """
    Base database connector for aladindb
    """

    type_id_to_cache_mapping = {
        'model': 'models',
        'metric': 'metrics',
        'encoder': 'encoders',
        'vector_index': 'vector_indices',
    }

    def __init__(
        self,
        databackend: BaseDataBackend,
        metadata: MetaDataStore,
        artifact_store: ArtifactStore,
        compute: ComputeBackend = LocalComputeBackend(),
    ):
        """
        :param databackend: object containing connection to Datastore
        :param metadata: object containing connection to Metadatastore
        :param artifact_store: object containing connection to Artifactstore
        :param compute: object containing connection to ComputeBackend
        """
        logging.info("Building Data Layer")

        self.metrics = LoadDict(self, field='metric')
        self.models = LoadDict(self, field='model')
        self.encoders = LoadDict(self, field='encoder')
        self.vector_indices = LoadDict(self, field='vector_index')
        self.serializers = LoadDict(self, field='serializer')
        for ser in serializers:
            self.serializers[ser] = serializers[ser]

        self.fast_vector_searchers = LoadDict(
            self, callable=self.initialize_vector_searcher
        )
        self.metadata = metadata
        self.artifact_store = artifact_store
        self.artifact_store.serializers = self.serializers

        self.databackend = databackend
        # TODO: force set config stores connection url

        self.cdc = DatabaseChangeDataCapture(self)

        self.compute = compute
        self._server_mode = False

    def rebuild(self, cfg=None):
        from aladindb.base import build

        self.databackend = build.build_databackend(cfg.data_backend if cfg else None)
        self.compute = build.build_compute(cfg.cluster.compute if cfg else None)

        if cfg:
            self.metadata = build.build_metadata(cfg.metadata_store)
            self.artifact_store = build.build_artifact_store(cfg.artifact_store)
            self.artifact_store.serializers = self.serializers
        else:
            self.metadata = self.databackend.build_metadata()
            self.artifact_store = self.databackend.build_artifact_store()
            self.artifact_store.serializers = self.serializers

    @property
    def server_mode(self):
        return self._server_mode

    @server_mode.setter
    def server_mode(self, is_server: bool):
        assert isinstance(is_server, bool)
        self._server_mode = is_server

    def initialize_vector_searcher(
        self, identifier, searcher_type: t.Optional[str] = None, backfill=False
    ) -> t.Optional[BaseVectorSearcher]:
        searcher_type = searcher_type or s.CFG.cluster.vector_search_type
        logging.info(f"loading of vectors of vector-index: '{identifier}'")
        vi = self.vector_indices[identifier]

        clt = vi.indexing_listener.select.table_or_collection

        if self.cdc.running:
            msg = 'CDC only supported for vector search via lance format'
            assert s.CFG.cluster.vector_search_type == 'lance', msg

        vector_search_cls = vector_searcher_implementations[searcher_type]
        vector_comparison = vector_search_cls.from_component(vi)

        assert isinstance(clt.identifier, str), 'clt.identifier must be a string'

        if backfill or not s.CFG.cluster.is_remote_vector_search:
            self.backfill_vector_search(vi, vector_comparison)

        return FastVectorSearcher(self, vector_comparison, vi.identifier)

    def backfill_vector_search(self, vi, searcher):
        if vi.indexing_listener.select is None:
            raise ValueError('.select must be set')

        key = vi.indexing_listener.key
        if key.startswith('_outputs.'):
            key = key.split('.')[1]

        model_id = vi.indexing_listener.model.identifier
        model_version = vi.indexing_listener.model.version

        query = vi.indexing_listener.select.outputs(
            **{key: f'{model_id}/{model_version}'}
        )

        logging.info(str(query))

        id_field = query.table_or_collection.primary_id

        progress = tqdm.tqdm(desc='Loading vectors into vector-table...')
        for record_batch in ibatch(
            self.execute(query),
            s.CFG.cluster.backfill_batch_size,
        ):
            items = []
            for record in record_batch:
                key = vi.indexing_listener.key
                if key.startswith('_outputs.'):
                    key = key.split('.')[1]

                id = record[id_field]
                assert not isinstance(vi.indexing_listener.model, str)
                h = record.outputs(
                    key,
                    vi.indexing_listener.model.identifier,
                    version=vi.indexing_listener.model.version,
                )
                if isinstance(h, Encodable):
                    h = h.x
                items.append(VectorItem.create(id=str(id), vector=h))

            searcher.add(items)
            progress.update(len(items))

    def set_compute(self, new: ComputeBackend):
        """
        Set a new compute engine at runtime. Use it only if you know what you do.
        The standard procedure is to set compute engine during initialization.
        """
        logging.warn(
            f"Change compute engine from '{self.compute.name}' to '{new.name}'"
        )

        self.compute.disconnect()
        logging.success(
            f"Succesfully disconnected from compute engine: '{self.compute.name}'"
        )

        logging.info(f"Connecting to compute engine: {new.name}")
        self.compute = new

    def get_compute(self):
        return self.compute

    def drop(self, force: bool = False):
        """
        Drop all data, artifacts and metadata
        """
        if not force and not click.confirm(
            f'{Colors.RED}[!!!WARNING USE WITH CAUTION AS YOU '
            f'WILL LOSE ALL DATA!!!]{Colors.RESET} '
            'Are you sure you want to drop the database? ',
            default=False,
        ):
            logging.warn('Aborting...')

        self.databackend.drop(force=True)
        self.metadata.drop(force=True)
        self.artifact_store.drop(force=True)

    def validate(
        self,
        identifier: str,
        type_id: str,
        validation_set: str,
        metrics: t.Sequence[str],
    ):
        """
        Evaluate quality of component, using `Component.validate`, if implemented.

        :param identifier: identifier of semantic index
        :param type_id: type_id of component
        :param validation_set: validation dataset on which to validate
        :param metrics: metric functions to compute
        """
        # TODO: never called
        component = self.load(type_id, identifier)
        metric_list = [self.load('metric', m) for m in metrics]
        assert isinstance(component, Model)
        return component.validate(
            self,
            validation_set,
            metric_list,  # type: ignore[arg-type]
        )

    def show(
        self,
        type_id: str,
        identifier: t.Optional[str] = None,
        version: t.Optional[int] = None,
    ):
        """
        Show available functionality which has been added using ``self.add``.
        If version is specified, then print full metadata

        :param type_id: type_id of component to show ['encoder', 'model', 'listener',
                       'learning_task', 'training_configuration', 'metric',
                       'vector_index', 'job']
        :param identifier: identifying string to component
        :param version: (optional) numerical version - specify for full metadata
        """
        if identifier is None and version is not None:
            raise ValueError(f'must specify {identifier} to go with {version}')

        if identifier is None:
            return self.metadata.show_components(type_id=type_id)

        if version is None:
            return sorted(
                self.metadata.show_component_versions(
                    type_id=type_id, identifier=identifier
                )
            )

        if version == -1:
            return self._get_object_info(type_id=type_id, identifier=identifier)

        return self._get_object_info(
            type_id=type_id, identifier=identifier, version=version
        )

    def _get_context(
        self, model, context_select: t.Optional[Select], context_key: t.Optional[str]
    ):
        assert model.takes_context, 'model does not take context'
        assert context_select is not None
        sources = list(self.execute(context_select))
        context = [x.unpack() for x in sources]

        if context_key is not None:
            context = [MongoStyleDict(x)[context_key] for x in context]
        return context, sources

    async def apredict(
        self,
        model_name: str,
        input: t.Union[Document, t.Any],
        context_select: t.Optional[Select] = None,
        context_key: t.Optional[str] = None,
        **kwargs,
    ):
        """
        Apply model to input using asyncio.

        :param model_name: model identifier
        :param input: input to be passed to the model.
                      Must be possible to encode with registered encoders
        :param context_select: select query object to provide context
        :param context_key: key to use to extract context from context_select
        """
        model = self.models[model_name]
        context = None
        sources: t.List[Document] = []

        if context_select is not None:
            context, sources = self._get_context(model, context_select, context_key)
        out = await model.apredict(
            input.unpack() if isinstance(input, Document) else input,
            one=True,
            context=context,
            **kwargs,
        )

        if model.encoder is not None:
            out = model.encoder(out)

        if context is not None:
            return Document(out), sources
        return Document(out), []

    def predict(
        self,
        model_name: str,
        input: t.Union[Document, t.Any],
        context_select: t.Optional[t.Union[str, Select]] = None,
        context_key: t.Optional[str] = None,
        **kwargs,
    ) -> t.Tuple[Document, t.List[Document]]:
        """
        Apply model to input.

        :param model_name: model identifier
        :param input: input to be passed to the model.
                      Must be possible to encode with registered encoders
        :param context_select: select query object to provide context
        :param context_key: key to use to extract context from context_select
        """
        model = self.models[model_name]
        context = None
        sources: t.List[Document] = []

        if context_select is not None:
            if isinstance(context_select, Select):
                context, sources = self._get_context(model, context_select, context_key)
            elif isinstance(context_select, str):
                context = context_select
            else:
                raise TypeError("context_select should be either Select or str")

        out = model.predict(
            input.unpack() if isinstance(input, Document) else input,
            one=True,
            context=context,
            **kwargs,
        )

        if isinstance(model.encoder, Encoder):
            out = model.encoder(out)

        if context is not None:
            return Document(out), sources
        return Document(out), []

    def execute(self, query: ExecuteQuery, *args, **kwargs) -> ExecuteResult:
        """
        Execute a query on the db.

        :param query: select, insert, delete, update,
        """

        if isinstance(query, Delete):
            return self.delete(query, *args, **kwargs)
        if isinstance(query, Insert):
            return self.insert(query, *args, **kwargs)
        if isinstance(query, Select):
            return self.select(query, *args, **kwargs)
        if isinstance(query, Table):
            return self.select(query.to_query(), *args, **kwargs)
        if isinstance(query, Update):
            return self.update(query, *args, **kwargs)
        if isinstance(query, RawQuery):
            return query.execute(self)

        raise TypeError(
            f'Wrong type of {query}; '
            f'Expected object of type {t.Union[Select, Delete, Update, Insert, str]}; '
            f'Got {type(query)};'
        )

    def delete(self, delete: Delete, refresh: bool = True) -> DeleteResult:
        """
        Delete data.

        :param delete: delete query object
        """
        result = delete.execute(self)
        if refresh and not self.cdc.running:
            return result, self.refresh_after_delete(delete, ids=result)
        return result, None

    def insert(
        self, insert: Insert, refresh: bool = True, encoders: t.Sequence[Encoder] = ()
    ) -> InsertResult:
        """
        Insert data.

        :param insert: insert query object
        """
        for e in encoders:
            self.add(e)
        for r in insert.documents:
            r['_fold'] = 'train'  # type: ignore[assignment]
            if random.random() < s.CFG.fold_probability:
                r['_fold'] = 'valid'  # type: ignore[assignment]
        inserted_ids = insert.execute(self)

        if refresh and self.cdc.running:
            raise Exception('cdc cannot be activated and refresh=True')

        if s.CFG.cluster.cdc is not None:
            logging.info('CDC active, skipping refresh')
            return inserted_ids, None

        if refresh:
            return inserted_ids, self.refresh_after_update_or_insert(
                insert, ids=inserted_ids, verbose=False
            )
        return inserted_ids, None

    def select(self, select: Select) -> SelectResult:
        """
        Select data.

        :param select: select query object
        """
        if select.variables:
            select = select.set_variables(self)  # type: ignore[assignment]
        return select.execute(self)

    def refresh_after_delete(
        self,
        query: Delete,
        ids: t.Sequence[str],
        verbose: bool = False,
    ):
        """
        Trigger cleanup jobs after data deletion.

        :param query: Select or Update which reduces scope of computations
        :param ids: ids which reduce scopy of computations
        :param verbose: Toggle to ``True`` to get more output
        """
        task_workflow: TaskWorkflow = self._build_delete_task_workflow(
            query,
            ids=ids,
            verbose=verbose,
        )
        task_workflow.run_jobs()
        return task_workflow

    def refresh_after_update_or_insert(
        self,
        query: t.Union[Insert, Select, Update],
        ids: t.Sequence[str],
        verbose: bool = False,
    ):
        """
        Trigger computation jobs after data insertion.

        :param query: Select or Update which reduces scope of computations
        :param ids: ids which reduce scopy of computations
        :param verbose: Toggle to ``True`` to get more output
        """
        task_workflow: TaskWorkflow = self._build_task_workflow(
            query.select_table,  # TODO can be replaced by select_using_ids
            ids=ids,
            verbose=verbose,
        )
        task_workflow.run_jobs()
        return task_workflow

    def update(self, update: Update, refresh: bool = True) -> UpdateResult:
        """
        Update data.

        :param update: update query object
        """
        updated_ids = update.execute(self)

        if refresh and self.cdc.running:
            raise Exception('cdc cannot be activated and refresh=True')
        if refresh:
            return updated_ids, self.refresh_after_update_or_insert(
                query=update, ids=updated_ids, verbose=False
            )
        return updated_ids, None

    def add(
        self,
        object: t.Union[Component, t.Sequence[t.Any], t.Any],
        dependencies: t.Sequence[Job] = (),
    ):
        """
        Add functionality in the form of components. Components are stored in the
        configured artifact store, and linked to the primary db through
        the metadata.

        :param object: Object to be stored
        :param dependencies: list of jobs which should execute before component
                             init begins
        """
        if isinstance(object, (list, tuple)):
            return type(object)(
                self._add(
                    object=component,
                    dependencies=dependencies,
                )
                for component in object
            )
        elif isinstance(object, Component):
            return self._add(object=object, dependencies=dependencies), object
        else:
            return self._add(aladin(object)), object

    def remove(
        self,
        type_id: str,
        identifier: str,
        version: t.Optional[int] = None,
        force: bool = False,
    ):
        """
        Remove component (version: optional)

        :param type_id: type_id of component to remove ['encoder', 'model', 'listener',
                        'training_configuration', 'vector_index']
        :param identifier: identifier of component (see `container.base.Component`)
        :param version: [optional] numerical version to remove
        :param force: force skip confirmation (use with caution)
        """
        # TODO: versions = [version] if version is not None else ...
        if version is not None:
            return self._remove_component_version(
                type_id, identifier, version=version, force=force
            )
        versions = self.metadata.show_component_versions(type_id, identifier)
        versions_in_use = []
        for v in versions:
            if self.metadata.component_version_has_parents(type_id, identifier, v):
                versions_in_use.append(v)

        if versions_in_use:
            component_versions_in_use = []
            for v in versions_in_use:
                unique_id = Component.make_unique_id(type_id, identifier, v)
                component_versions_in_use.append(
                    f"{unique_id} -> "
                    f"{self.metadata.get_component_version_parents(unique_id)}",
                )
            if not force:
                raise exceptions.ComponentInUseError(
                    f'Component versions: {component_versions_in_use} are in use'
                )
            else:
                warnings.warn(
                    exceptions.ComponentInUseWarning(
                        f'Component versions: {component_versions_in_use}'
                        ', marking as hidden'
                    )
                )

        if force or click.confirm(
            f'You are about to delete {type_id}/{identifier}, are you sure?',
            default=False,
        ):
            for v in sorted(list(set(versions) - set(versions_in_use))):
                self._remove_component_version(type_id, identifier, v, force=True)

            for v in sorted(versions_in_use):
                self.metadata.hide_component_version(type_id, identifier, v)
        else:
            logging.warn('aborting.')

    def load(
        self,
        type_id: str,
        identifier: str,
        version: t.Optional[int] = None,
        allow_hidden: bool = False,
        info_only: bool = False,
    ) -> t.Union[Component, t.Dict[str, t.Any]]:
        """
        Load component using uniquely identifying information.

        :param type_id: type_id of component to remove
                        ['encoder', 'model', 'listener', ...]
        :param identifier: identifier of component (see `container.base.Component`)
        :param version: [optional] numerical version
        :param allow_hidden: toggle to ``True`` to allow loading of deprecated
                             components
        :param info_only: toggle to ``True`` to return metadata only
        """
        info = self.metadata.get_component(
            type_id=type_id,
            identifier=identifier,
            version=version,
            allow_hidden=allow_hidden,
        )

        if info is None:
            raise exceptions.MetadataException(
                f'No such object of type "{type_id}", '
                f'"{identifier}" has been registered.'
            )

        if info_only:
            return info

        def get_children(info):
            return {
                k: v
                for k, v in info['dict'].items()
                if isinstance(v, dict) and serializable.is_component_metadata(v)
            }

        def replace_children(r):
            if isinstance(r, dict):
                children = get_children(r)
                for k, v in children.items():
                    c = replace_children(
                        self.metadata.get_component(**v, allow_hidden=True)
                    )
                    try:
                        r['dict'][k] = c
                    except KeyError:
                        raise exceptions.MetadataException(
                            'Children {k} not found in `dict`'
                        )
            return r

        info = replace_children(info)
        info = self.artifact_store.load(info, lazy=True)

        m = Component.deserialize(info)
        m.db = self
        m.on_load(self)

        if cm := self.type_id_to_cache_mapping.get(type_id):
            try:
                getattr(self, cm)[m.identifier] = m
            except KeyError:
                raise exceptions.ComponentException('%s not found in %s cache'.format())
        return m

    def _build_delete_task_workflow(
        self,
        query: Delete,
        ids: t.Sequence[str],
        verbose: bool = False,
    ):
        G = TaskWorkflow(self)
        vector_indices = self.show('vector_index')

        if not vector_indices:
            return G

        deleted_table_or_collection = query.table_or_collection.identifier

        for vi in vector_indices:
            vi = self.vector_indices[vi]
            listener_table_or_collection = (
                vi.indexing_listener.select.table_or_collection.identifier
            )

            if deleted_table_or_collection != listener_table_or_collection:
                continue

            if (
                s.CFG.cluster.vector_search_type == 'in_memory'
                and vi.identifier not in self.fast_vector_searchers
            ):
                continue

            G.add_node(
                f'{vi.identifier}.{delete_vectors.__name__}()',
                job=FunctionJob(
                    callable=delete_vectors,
                    args=[],
                    kwargs=dict(
                        vector_index=vi.identifier,
                        ids=ids,
                    ),
                ),
            )

        return G

    def _build_task_workflow(
        self,
        query,
        ids=None,
        dependencies=(),
        verbose: bool = True,
    ) -> TaskWorkflow:
        logging.debug(f"Building task workflow graph. Query:{query}")

        job_ids: t.Dict[str, t.Any] = defaultdict(lambda: [])
        job_ids.update(dependencies)
        # TODO use these job_ids as dependencies for every job

        G = TaskWorkflow(self)

        # TODO extract this logic from this class
        G.add_node(
            f'{download_content.__name__}()',
            job=FunctionJob(
                callable=download_content,
                args=[],
                kwargs=dict(
                    ids=ids,
                    query=query.serialize(),
                ),
            ),
        )

        listeners = self.show('listener')
        if not listeners:
            return G
        listener_selects = {}

        for identifier in listeners:
            info = self.metadata.get_component('listener', identifier)
            listener_query = info['dict']['select']

            listener_select = serializable.Serializable.deserialize(listener_query)
            listener_selects.update({identifier: listener_select})
            if listener_select is None:
                continue
            if (
                listener_select.table_or_collection.identifier
                != query.table_or_collection.identifier
            ):
                continue

            model, _, key = identifier.rpartition('/')
            G.add_node(
                f'{model}.predict({key})',
                job=ComponentJob(
                    component_identifier=model,
                    args=[key],
                    kwargs={
                        'ids': ids,
                        'select': listener_query,
                        **info['dict']['predict_kwargs'],
                    },
                    method_name='predict',
                    type_id='model',
                ),
            )

        for identifier in listeners:
            listener_select = listener_selects[identifier]
            if listener_select is None:
                continue
            if (
                listener_select.table_or_collection.identifier
                != query.table_or_collection.identifier
            ):
                continue
            model, _, key = identifier.rpartition('/')
            G.add_edge(
                f'{download_content.__name__}()',
                f'{model}.predict({key})',
            )
            deps = self._get_dependencies_for_listener(identifier)
            for dep in deps:
                dep_model, _, dep_key = dep.rpartition('/')
                G.add_edge(
                    f'{dep_model}.predict({dep_key})',
                    f'{model}.predict({key})',
                )

        if s.CFG.self_hosted_vector_search:
            return G

        for identifier in self.show('vector_index'):
            # if a vector-searcher is not loaded, then skip
            # since s.CFG.cluster.vector_search_type == 'in_memory' implies the
            # program is standalone

            if (
                s.CFG.cluster.vector_search_type == 'in_memory'
                and identifier not in self.fast_vector_searchers
            ):
                continue

            vi = self.vector_indices[identifier]
            if (
                vi.indexing_listener.select.table_or_collection.identifier
                != query.table_or_collection.identifier
            ):
                continue

            G.add_node(
                f'{identifier}.{copy_vectors.__name__}',
                FunctionJob(
                    callable=copy_vectors,
                    args=[],
                    kwargs={
                        'vector_index': identifier,
                        'ids': ids,
                        'query': vi.indexing_listener.select.serialize(),
                    },
                ),
            )
            model = vi.indexing_listener.model.identifier
            key = vi.indexing_listener.key
            G.add_edge(
                f'{model}.predict({key})', f'{identifier}.{copy_vectors.__name__}'
            )

        return G

    def _compute_model_outputs(
        self,
        model_info,
        _ids,
        select: Select,
        key='_base',
        model=None,
        predict_kwargs=None,
    ):
        s.logging.info('finding documents under filter')
        model_identifier = model_info['identifier']
        documents = list(self.execute(select.select_using_ids(_ids)))
        documents = [x.unpack() for x in documents]
        if key != '_base':
            passed_docs = [r[key] for r in documents]
        else:
            passed_docs = documents
        if model is None:
            model = self.models[model_identifier]
        return model.predict(passed_docs, **(predict_kwargs or {}))

    def _save_artifacts(self, artifact_dictionary: t.Dict):
        raise NotImplementedError  # TODO use in the server code...

    def _add(
        self,
        object: Component,
        dependencies: t.Sequence[Job] = (),
        serialized: t.Optional[t.Dict] = None,
        parent: t.Optional[str] = None,
    ):
        jobs = []
        object.db = self
        object.pre_create(self)
        assert hasattr(object, 'identifier')
        assert hasattr(object, 'version')

        existing_versions = self.show(object.type_id, object.identifier)
        if isinstance(object.version, int) and object.version in existing_versions:
            s.logging.debug(f'{object.unique_id} already exists - doing nothing')
            return []

        if existing_versions:
            object.version = max(existing_versions) + 1
        else:
            object.version = 0

        if serialized is None:
            serialized, artifacts = object.serialized
            artifact_info = self.artifact_store.save(artifacts)
            serialized = self.artifact_store.replace(serialized, artifact_info)

        else:
            try:
                serialized['version'] = object.version
                serialized['dict']['version'] = object.version
            except KeyError:
                raise exceptions.MetadataException(
                    '`dict` or `version` not found in serialized dict.'
                )

        jobs.extend(self._create_children(object, serialized))
        self.metadata.create_component(serialized)
        if parent is not None:
            self.metadata.create_parent_child(parent, object.unique_id)
        object.post_create(self)
        self._add_component_to_cache(object)
        these_jobs = object.schedule_jobs(self, dependencies=dependencies)
        jobs.extend(these_jobs)
        return jobs

    def _create_children(self, component: Component, serialized: t.Dict):
        jobs = []
        for k, child_type_id in component.child_components:
            assert isinstance(k, str)
            child = getattr(component, k)
            if isinstance(child, str):
                serialized_dict = {
                    'type_id': child_type_id,
                    'identifier': child,
                    'version': self.metadata.get_latest_version(child_type_id, child),
                }
                self.metadata.create_parent_child(
                    component.unique_id, Component.make_unique_id(**serialized_dict)
                )
                serialized['dict'][k] = serialized_dict
            elif isinstance(child, Component):
                sub_jobs = self._add(
                    child,
                    serialized=serialized['dict'][k],
                    parent=component.unique_id,
                )
                jobs.extend(sub_jobs)
                serialized_dict = {
                    'type_id': child.type_id,
                    'identifier': child.identifier,
                    'version': child.version,
                }
                serialized['dict'][k] = serialized_dict
            elif isinstance(child, list):
                serialized_list = []
                for i, sub in enumerate(child):
                    assert isinstance(sub, Component)
                    sub_jobs = self._add(
                        sub,
                        serialized=serialized['dict'][k][i],
                        parent=component.unique_id,
                    )
                    jobs.extend(sub_jobs)
                    serialized_list.append(
                        {
                            'type_id': sub.type_id,
                            'identifier': sub.identifier,
                            'version': sub.version,
                        }
                    )
                serialized['dict'][k] = serialized_list
            else:
                raise ValueError(f'Unknown child type: {type(child)}')
        return jobs

    def _create_plan(self):
        G = networkx.DiGraph()
        for identifier in self.metadata.show_components('listener', active=True):
            G.add_node('listener', job=identifier)
        for identifier in self.metadata.show_components('listener'):
            deps = self._get_dependencies_for_listener(identifier)
            for dep in deps:
                G.add_edge(('listener', dep), ('listener', identifier))
        if not networkx.is_directed_acyclic_graph(G):
            raise ValueError('G is not a directed, acyclic graph')
        return G

    def _delete(self, delete: Delete):
        delete.execute(self)

    def _remove_component_version(
        self,
        type_id: str,
        identifier: str,
        version: int,
        force: bool = False,
    ):
        unique_id = Component.make_unique_id(type_id, identifier, version)
        if self.metadata.component_version_has_parents(type_id, identifier, version):
            parents = self.metadata.get_component_version_parents(unique_id)
            raise Exception(f'{unique_id} is involved in other components: {parents}')

        if force or click.confirm(
            f'You are about to delete {unique_id}, are you sure?',
            default=False,
        ):
            component = self.load(type_id, identifier, version=version)
            info = self.metadata.get_component(type_id, identifier, version=version)
            if hasattr(component, 'cleanup'):
                component.cleanup(self)
            if type_id in self.type_id_to_cache_mapping:
                try:
                    del getattr(self, self.type_id_to_cache_mapping[type_id])[
                        identifier
                    ]
                except KeyError:
                    pass

            if hasattr(component, 'artifact_attributes'):
                for a in component.artifact_attributes:
                    if info['dict'][a] is not None:
                        self.artifact_store.delete(info['dict'][a]['file_id'])
            self.metadata.delete_component_version(type_id, identifier, version=version)

    def _download_content(  # TODO: duplicated function
        self,
        query: t.Optional[t.Union[Select, Insert]] = None,
        ids=None,
        documents=None,
        timeout=None,
        raises: bool = True,
        n_download_workers=None,
        headers=None,
        **kwargs,
    ):
        update_db = False

        if documents is not None:
            pass
        elif isinstance(query, Select):
            update_db = True
            if ids is None:
                documents = list(self.select(query))
            else:
                select = query.select_using_ids(ids)
                cursor = self.select(select).raw_cursor
                documents = [Document(x) for x in cursor]
        elif isinstance(query, Insert):
            documents = query.documents
        else:
            raise NotImplementedError

        documents = [x.content for x in documents]
        uris, keys, place_ids = gather_uris(documents)
        s.logging.info(f'found {len(uris)} uris')
        if not uris:
            return

        if n_download_workers is None:
            try:
                n_download_workers = self.metadata.get_metadata(
                    key='n_download_workers'
                )
            except exceptions.MetadatastoreException:
                n_download_workers = 0

        if headers is None:
            try:
                headers = self.metadata.get_metadata(key='headers')
            except exceptions.MetadatastoreException:
                headers = 0

        if timeout is None:
            try:
                timeout = self.metadata.get_metadata(key='download_timeout')
            except exceptions.MetadatastoreException:
                timeout = None

        def download_update(key, id, bytes):
            return query.download_update(db=self, key=key, id=id, bytes=bytes)

        downloader = Downloader(
            uris=uris,
            ids=place_ids,
            keys=keys,
            update_one=download_update,
            n_workers=n_download_workers,
            timeout=timeout,
            headers=headers,
            raises=raises,
        )
        downloader.go()
        if update_db:
            return
        for id_, key in zip(place_ids, keys):
            documents[id_] = self.databackend.set_content_bytes(
                documents[id_], key, downloader.results[id_]  # type: ignore[index]
            )
        return documents

    def _get_content_for_filter(self, filter) -> Document:
        if isinstance(filter, dict):
            filter = Document(content=filter)
        if '_id' not in filter.content:
            filter['_id'] = 0
        uris = gather_uris([filter.content])[0]
        if uris:
            output = self._download_content(
                query=None, documents=[filter.content], timeout=None, raises=True
            )[0]
            filter = Document(Document.decode(output, encoders=self.encoders))
        if not filter['_id']:
            del filter.content['_id']
        return filter

    def _get_dependencies_for_listener(self, identifier):
        info = self.metadata.get_component('listener', identifier)
        if info is None:
            return []
        out = []
        if info['dict']['key'].startswith('_outputs.'):
            _, key, model = info['dict']['key'].split('.')
            out.append(f'{model}/{key}')
        return out

    def _get_file_content(self, r):
        for k in r:
            if isinstance(r[k], dict):
                r[k] = self._get_file_content(r[k])
        return r

    def _get_object_info(self, identifier, type_id, version=None):
        return self.metadata.get_component(type_id, identifier, version=version)

    def _apply_listener(
        self,
        identifier,
        ids: t.Optional[t.Sequence[str]] = None,
        verbose: bool = False,
        max_chunk_size=5000,
        model=None,
        recompute: bool = False,
        listener_info=None,
        **kwargs,
    ) -> t.List:
        # NOTE: this method is never called anywhere except for itself!
        if listener_info is None:
            listener_info = self.metadata.get_component('listener', identifier)

        select = serializable.Serializable.deserialize(listener_info['select'])
        if ids is None:
            ids = select.get_ids(self)
        else:
            ids = select.select_using_ids(ids=ids).get_ids(self)
            assert ids is not None

        ids = [str(id) for id in ids]

        if max_chunk_size is not None:
            for it, i in enumerate(range(0, len(ids), max_chunk_size)):
                s.logging.info(
                    'computing chunk '
                    f'({it + 1}/{math.ceil(len(ids) / max_chunk_size)})'
                )
                self._apply_listener(
                    identifier,
                    ids=ids[i : i + max_chunk_size],
                    verbose=verbose,
                    max_chunk_size=None,
                    model=model,
                    recompute=recompute,
                    listener_info=listener_info,
                    **kwargs,
                )
            return []

        model_info = self.metadata.get_component('model', listener_info['model'])
        outputs = self._compute_model_outputs(
            model_info,
            ids,
            select,
            key=listener_info['key'],
            model=model,
            predict_kwargs=listener_info.get('predict_kwargs', {}),
        )
        type = model_info.get('type')
        if type is not None:
            type = self.encoders[type]
            outputs = [type(x).encode() for x in outputs]

        select.model_update(
            db=self,
            model=listener_info['model'],
            key=listener_info['key'],
            version=listener_info['version'],
            outputs=outputs,
            ids=ids,
        )
        return outputs

    def replace(
        self,
        object: t.Any,
        upsert: bool = False,
    ):
        """
        (Use-with caution!!) Replace a model in artifact store with updated object.
        :param object: object to replace
        :param upsert: toggle to ``True`` to enable even if object doesn't exist yet
        """
        try:
            info = self.metadata.get_component(
                object.type_id, object.identifier, version=object.version
            )
        except FileNotFoundError:
            if upsert:
                return self.add(
                    object,
                )
            raise FileNotFoundError

        # If object has no version, update the last version
        object.version = info['version']
        new_info = self.artifact_store.update(object, metadata_info=info)
        self.metadata.replace_object(
            new_info,
            identifier=object.identifier,
            type_id='model',
            version=object.version,
        )

    def select_nearest(
        self,
        like: t.Union[t.Dict, Document],
        vector_index: str,
        ids: t.Optional[t.Sequence[str]] = None,
        outputs: t.Optional[Document] = None,
        n: int = 100,
    ) -> t.Tuple[t.List[str], t.List[float]]:
        if not isinstance(like, Document):
            assert isinstance(like, dict)
            like = Document(like)
        like = self._get_content_for_filter(like)
        vi = self.vector_indices[vector_index]
        if outputs is None:
            outs = {}
        else:
            outs = outputs.encode()
            if not isinstance(outs, dict):
                raise TypeError(f'Expected dict, got {type(outputs)}')
        return vi.get_nearest(like, db=self, ids=ids, n=n, outputs=outs)

    def close(self):
        """
        Gracefully shutdown the Datalayer
        """
        logging.info("Disconnect from Data Store")
        self.databackend.disconnect()

        logging.info("Disconnect from Metadata Store")
        self.metadata.disconnect()

        logging.info("Disconnect from Artifact Store")
        self.artifact_store.disconnect()

        logging.info("Disconnect from Compute Engine")
        self.compute.disconnect()

        # TODO: gracefully close all opened connections
        return

    def _add_component_to_cache(self, component: Component):
        """
        Add component to cache when it is added to the db.
        Avoiding the need to load it from the db again.
        """
        type_id = component.type_id
        if cm := self.type_id_to_cache_mapping.get(type_id):
            getattr(self, cm)[component.identifier] = component
        component.on_load(self)


@dc.dataclass
class LoadDict(dict):
    database: Datalayer
    field: t.Optional[str] = None
    callable: t.Optional[t.Callable] = None

    def __missing__(self, key: str):
        if self.field is not None:
            value = self[key] = self.database.load(self.field, key)
        else:
            msg = f'callable is ``None`` for {key}'
            assert self.callable is not None, msg
            value = self[key] = self.callable(key)
        return value



import base64
import os
import typing as t

import numpy as np

if t.TYPE_CHECKING:
    from aladindb.components.encoder import Encoder


def str_shape(shape: t.Sequence[int]) -> str:
    if not shape:
        raise ValueError('Shape was empty')
    return 'x'.join(str(x) for x in shape)


def get_key(key_name: str) -> str:
    try:
        return os.environ[key_name]
    except KeyError:
        raise KeyError(f'Environment variable {key_name} is not set') from None


def format_prompt(X: str, prompt: str, context: t.Optional[t.List[str]] = None) -> str:
    format_params = {}
    if '{input}' in prompt:
        format_params['input'] = X
    else:
        prompt += X

    if '{context}' in prompt:
        if context:
            format_params['context'] = '\n'.join(context)
        else:
            raise ValueError(f'A context is required for prompt {prompt}')

    return prompt.format(**format_params)


def aladinencode(object):
    if isinstance(object, np.ndarray):
        from aladindb.ext.numpy import array

        encoded = array(dtype=object.dtype, shape=object.shape).encode(object)
        encoded['shape'] = object.shape
        encoded['dtype'] = str(object.dtype)
        return encoded
    return object


def aladindecode(r: t.Any, encoders: t.List['Encoder']):
    if isinstance(r, dict):
        encoder = encoders[r['_content']['encoder']]
        b = base64.b64decode(r['_content']['bytes'])
        return encoder.decode(b).x
    return r



import re
import typing as t

from aladindb.base.configs import CFG

__all__ = ('aladin',)


def aladin(item: t.Optional[t.Any] = None, **kwargs) -> t.Any:
    """
    Attempts to automatically wrap an item in a aladindb component by
    using duck typing to recognize it.

    :param item: A database or model
    """

    if item is None:
        from aladindb.base.build import build_datalayer

        return build_datalayer()

    if isinstance(item, str):
        return _auto_identify_connection_string(item, **kwargs)

    return _DuckTyper.run(item, **kwargs)


def _auto_identify_connection_string(item: str, **kwargs) -> t.Any:
    from aladindb.base.build import build_datalayer

    if item.startswith('mongomock://'):
        CFG.force_set('data_backend', item)

    elif item.startswith('mongodb://'):
        CFG.force_set('data_backend', item)

    elif item.startswith('mongodb+srv://') and 'mongodb.net' in item:
        CFG.force_set('data_backend', item)
        CFG.force_set('cluster.vector_search', item)

    elif item.endswith('.csv'):
        CFG.force_set('data_backend', item)

    else:
        if re.match(r'^[a-zA-Z0-9]+://', item) is None:
            raise ValueError(f'{item} is not a valid connection string')
        CFG.force_set('data_backend', item)
    return build_datalayer(CFG, **kwargs)


class _DuckTyper:
    attrs: t.Sequence[str]
    count: int

    @staticmethod
    def run(item: t.Any, **kwargs) -> t.Any:
        dts = [dt for dt in _DuckTyper._DUCK_TYPES if dt.accept(item)]
        if not dts:
            raise ValueError(
                f'Couldn\'t auto-identify {item}, please wrap explicitly using '
                '``aladindb.components.*``'
            )

        if len(dts) == 1:
            return dts[0].create(item, **kwargs)

        raise ValueError(f'{item} matched more than one type: {dts}')

    @classmethod
    def accept(cls, item: t.Any) -> bool:
        """Does this item match the DuckType?

        The default implementation returns True if the number of attrs that
        the item has is exactly equal to self.count.

        """
        return sum(hasattr(item, a) for a in cls.attrs) == cls.count

    @classmethod
    def create(cls, item: t.Any, **kwargs) -> t.Any:
        """Create a aladindb component for an item that has already been accepted"""
        raise NotImplementedError

    _DUCK_TYPES: t.List[t.Type] = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        _DuckTyper._DUCK_TYPES.append(cls)


class MongoDbTyper(_DuckTyper):
    attrs = ('list_collection_names',)
    count = len(attrs)

    @classmethod
    def accept(cls, item: t.Any) -> bool:
        return super().accept(item) and item.__class__.__name__ == 'Database'

    @classmethod
    def create(cls, item: t.Any, **kwargs) -> t.Any:
        from mongomock.database import Database as MockDatabase
        from pymongo.database import Database

        from aladindb import logging
        from aladindb.backends.mongodb.data_backend import MongoDataBackend
        from aladindb.base.build import build_datalayer

        if not isinstance(item, (Database, MockDatabase)):
            raise TypeError(f'Expected Database but got {type(item)}')

        logging.warn(
            'Note: This is only recommended in development mode, since config\
             still holds `data_backend` with the default value, services \
             like vector search and cdc cannot be reached due to configuration\
             mismatch. Services will be configured with a `data_backend` uri using \
             config file hence this client config and\
             services config will be different.'
        )
        databackend = MongoDataBackend(conn=item.client, name=item.name)
        return build_datalayer(cfg=CFG, databackend=databackend, **kwargs)


class SklearnTyper(_DuckTyper):
    attrs = '_predict', 'fit', 'score', 'transform'
    count = 2

    @classmethod
    def create(cls, item: t.Any, **kwargs) -> t.Any:
        from sklearn.base import BaseEstimator

        from aladindb.ext.sklearn.model import Estimator

        if not isinstance(item, BaseEstimator):
            raise TypeError('Expected BaseEstimator but got {type(item)}')

        kwargs['identifier'] = _auto_identify(item)
        return Estimator(object=item, **kwargs)


class TorchTyper(_DuckTyper):
    attrs = 'forward', 'parameters', 'state_dict', '_load_from_state_dict'
    count = len(attrs)

    @classmethod
    def create(cls, item: t.Any, **kwargs) -> t.Any:
        from torch import jit, nn

        from aladindb.ext.torch.model import TorchModel

        if isinstance(item, nn.Module) or isinstance(item, jit.ScriptModule):
            return TorchModel(identifier=_auto_identify(item), object=item, **kwargs)

        raise TypeError('Expected a Module but got {type(item)}')


def _auto_identify(instance: t.Any) -> str:
    return instance.__class__.__name__.lower()



import itertools
import typing as t

T = t.TypeVar('T')


def ibatch(iterable: t.Iterable[T], batch_size: int) -> t.Iterator[t.List[T]]:
    """
    Batch an iterable into chunks of size `batch_size`

    :param iterable: the iterable to batch
    :param batch_size: the number of groups to write
    """

    iterator = iter(iterable)
    while True:
        batch = list(itertools.islice(iterator, batch_size))
        if not batch:
            break
        yield batch



import typing as t

from aladindb.backends.base.query import CompoundSelect
from aladindb.base.serializable import Serializable
from aladindb.misc.special_dicts import MongoStyleDict
from aladindb.vector_search.base import VectorItem


def delete_vectors(
    vector_index: str,
    ids: t.Sequence[str],
    db=None,
):
    """
    A helper fxn to delete vectors of a ``VectorIndex`` component
    in the fast_vector_search backend.

    :param vector_index: A identifier of vector-index.
    :param ids: List of ids which were observed as deleted documents.
    :param db: A ``DB`` instance.
    """
    return db.fast_vector_searchers[vector_index].delete(ids)


def copy_vectors(
    vector_index: str,
    query: t.Union[t.Dict, CompoundSelect],
    ids: t.Sequence[str],
    db=None,
):
    """
    A helper fxn to copy vectors of a ``VectorIndex`` component
    from the databackend to the fast_vector_search backend.

    :param vector-index: A identifier of the vector-index.
    :param query: A query which was used by `db._build_task_workflow` method
    :param ids: List of ids which were observed as added/updated documents.
    :param db: A ``DB`` instance.
    """
    vi = db.vector_indices[vector_index]
    if isinstance(query, dict):
        # ruff: noqa: E501
        query: CompoundSelect = Serializable.deserialize(query)  # type: ignore[no-redef]
    assert isinstance(query, CompoundSelect)
    select = query.select_using_ids(ids)
    docs = db.select(select)
    docs = [doc.unpack() for doc in docs]
    key = vi.indexing_listener.key
    model = vi.indexing_listener.model.identifier
    version = vi.indexing_listener.model.version
    vectors = [
        {
            'vector': MongoStyleDict(doc)[f'_outputs.{key}.{model}.{version}'],
            'id': str(doc['_id']),
        }
        for doc in docs
    ]
    for r in vectors:
        if hasattr(r['vector'], 'numpy'):
            r['vector'] = r['vector'].numpy()
    db.fast_vector_searchers[vi.identifier].add(
        [VectorItem(**vector) for vector in vectors]
    )



import copy
import dataclasses as dc
import typing as t

import mongomock
import pymongo
from bson import ObjectId
from pymongo import InsertOne as _InsertOne, UpdateOne as _UpdateOne

from aladindb.backends.base.query import (
    CompoundSelect,
    Delete,
    Insert,
    Like,
    QueryComponent,
    QueryLinker,
    QueryType,
    Select,
    TableOrCollection,
    Update,
)
from aladindb.base.cursor import aladinCursor
from aladindb.base.document import Document
from aladindb.base.serializable import Variable


class FindOne(QueryComponent):
    """
    Wrapper around ``pymongo.Collection.find_one``

    :param args: Positional arguments to ``pymongo.Collection.find_one``
    :param kwargs: Named arguments to ``pymongo.Collection.find_one``
    """

    def select_using_ids(self, ids):
        ids = [ObjectId(id) for id in ids]
        args = list(self.args)[:]
        if not args:
            args = [{}]
        args[0].update({'_id': {'$in': ids}})
        return FindOne(
            name=self.name,
            type=self.type,
            args=args,
            kwargs=self.kwargs,
        )

    def add_fold(self, fold: str):
        """
        Modify the query to add a fold to filter {'_fold': fold}

        :param fold: The fold to add
        """
        if not self.args:
            args: t.List[t.Any] = [{}]
        args[0]['_fold'] = fold
        return FindOne(
            name=self.name,
            type=self.type,
            args=args,
            kwargs=self.kwargs,
        )


@dc.dataclass
class Find(QueryComponent):
    """
    Wrapper around ``pymongo.Collection.find``

    :param args: Positional arguments to ``pymongo.Collection.find``
    :param kwargs: Named arguments to ``pymongo.Collection.find``
    """

    output_fields: t.Optional[t.Dict[str, str]] = None

    def __post_init__(self):
        if not self.args:
            self.args = [{}]
        else:
            self.args = list(self.args)
        if not self.args[1:]:
            self.args.append({})

        if self.output_fields is not None:
            # if outputs are specified, then project to those outputs
            self.args[1].update(
                {f'_outputs.{k}.{v}': 1 for k, v in self.output_fields.items()}
            )
            if '_id' not in self.args[1]:
                self.args[1]['_id'] = 1

    @property
    def select_ids(self):
        args = list(self.args)[:]
        if not args:
            args = [{}]
        else:
            args = list(self.args)
        if not args[1:]:
            args.append({})

        args[1].update({'_id': 1})
        return Find(
            name=self.name,
            type=self.type,
            args=args,
            kwargs=self.kwargs,
        )

    def outputs(self, **kwargs):
        """
        Join the query with the outputs for a table.

        :param **kwargs: key=model/version or key=model pairs
        """
        args = list(self.args[:])
        if not args:
            args = [{}]
        if not args[1:]:
            args.append({})

        for k, v in kwargs.items():
            if '/' in v:
                model, version = v.split('/')
                args[1][f'_outputs.{k}.{model}.{version}'] = 1
            else:
                args[1][
                    Variable(
                        f'_outputs.{k}.{v}' + '.{version}',
                        lambda db, value, kwargs: value.format(
                            version=db.show('model', model)[-1]
                        ),
                    )
                ] = 1
        return Find(
            name=self.name,
            type=self.type,
            args=self.args,
            kwargs=self.kwargs,
            output_fields=kwargs,
        )

    def select_using_ids(self, ids):
        ids = [ObjectId(id) for id in ids]
        args = list(self.args)[:]
        if not args:
            args = [{}]
        args[0].update({'_id': {'$in': ids}})
        return Find(
            name=self.name,
            type=self.type,
            args=args,
            kwargs=self.kwargs,
        )

    def select_ids_of_missing_outputs(self, key: str, model: str, version: int):
        assert self.type == QueryType.QUERY
        if self.args:
            args = [
                {
                    '$and': [
                        self.args[0],
                        {f'_outputs.{key}.{model}.{version}': {'$exists': 0}},
                    ]
                },
                *self.args[1],
            ]
        else:
            args = [{f'_outputs.{key}.{model}': {'$exists': 0}}]

        return Find(
            name='find',
            type=QueryType.QUERY,
            args=args,
            kwargs=self.kwargs,
        )

    def select_single_id(self, id):
        assert self.type == QueryType.QUERY
        args = list(self.args)[:]
        if not args:
            args = [{}]
        args[0].update({'_id': id})
        return QueryComponent(
            name='find_one',
            type=QueryType.QUERY,
            args=args,
            kwargs=self.kwargs,
        )

    def add_fold(self, fold: str):
        args = self.args
        if not self.args:
            args = [{}]
        args[0]['_fold'] = fold
        return FindOne(
            name=self.name,
            type=self.type,
            args=args,
            kwargs=self.kwargs,
        )


@dc.dataclass
class Aggregate(Select):
    """
    Wrapper around ``pymongo.Collection.aggregate``

    :param table_or_collection: The table or collection to perform the query on
    :param vector_index: The vector index to use
    :param args: Positional arguments to ``pymongo.Collection.aggregate``
    :param kwargs: Named arguments to ``pymongo.Collection.aggregate``
    """

    table_or_collection: 'TableOrCollection'
    vector_index: t.Optional[str] = None
    args: t.Tuple[t.Any, ...] = dc.field(default_factory=tuple)
    kwargs: t.Dict[str, t.Any] = dc.field(default_factory=dict)

    @property
    def id_field(self):
        return self.table_or_collection.primary_id

    @property
    def select_table(self):
        raise NotImplementedError

    def add_fold(self):
        raise NotImplementedError

    def select_single_id(self, id: str):
        raise NotImplementedError

    @property
    def select_ids(self):
        raise NotImplementedError

    def select_using_ids(self):
        raise NotImplementedError

    def select_ids_of_missing_outputs(self, key: str, model: str, version: int):
        raise NotImplementedError

    @staticmethod
    def _replace_document_with_vector(step, vector_index, db):
        step = copy.deepcopy(step)
        assert "like" in step['$vectorSearch']
        vector = step['$vectorSearch']['like']

        if not isinstance(vector, Document):
            vector_index = db.vector_indices[vector_index]
            models, keys = vector_index.models_keys
            vector, _, _ = vector_index.get_vector(
                like=vector,
                models=models,
                keys=keys,
                db=db,
            )

        step['$vectorSearch']['queryVector'] = vector
        indexing_key = vector_index.indexing_listener.key
        if indexing_key.startswith('_outputs'):
            indexing_key = indexing_key.split('.')[1]
        indexing_model = vector_index.indexing_listener.model.identifier
        indexing_version = vector_index.indexing_listener.model.version
        step['$vectorSearch'][
            'path'
        ] = f'_outputs.{indexing_key}.{indexing_model}.{indexing_version}'
        step['$vectorSearch']['index'] = vector_index.identifier
        del step['$vectorSearch']['like']
        return step

    @staticmethod
    def _prepare_pipeline(pipeline, db, vector_index):
        pipeline = copy.deepcopy(pipeline)
        try:
            search_step = next(
                (i, step) for i, step in enumerate(pipeline) if '$vectorSearch' in step
            )
        except StopIteration:
            return pipeline
        pipeline[search_step[0]] = Aggregate._replace_document_with_vector(
            search_step[1],
            vector_index,
            db,
        )
        return pipeline

    def execute(self, db):
        collection = db.databackend.get_table_or_collection(
            self.table_or_collection.identifier
        )
        cursor = collection.aggregate(
            self._prepare_pipeline(
                self.args[0],
                db,
                self.vector_index,
            )
        )
        return aladinCursor(
            raw_cursor=cursor,
            id_field='_id',
            encoders=db.encoders,
        )


@dc.dataclass(repr=False)
class MongoCompoundSelect(CompoundSelect):
    def _get_query_linker(self, table_or_collection, members) -> 'QueryLinker':
        return MongoQueryLinker(
            table_or_collection=table_or_collection, members=members
        )

    @property
    def output_fields(self):
        return self.query_linker.output_fields

    def outputs(self, **kwargs):
        """
        This method returns a query which joins a query with the outputs
        for a table.

        :param model: The model identifier for which to get the outputs

        >>> q = Collection(...).find(...).outputs('key', 'model_name')

        """
        assert self.query_linker is not None
        return MongoCompoundSelect(
            table_or_collection=self.table_or_collection,
            pre_like=self.pre_like,
            query_linker=self.query_linker.outputs(**kwargs),
            post_like=self.post_like,
        )

    def change_stream(self, *args, **kwargs):
        return self.table_or_collection.change_stream(*args, **kwargs)

    def _execute(self, db):
        similar_scores = None
        query_linker = self.query_linker
        if self.pre_like:
            similar_ids, similar_scores = self.pre_like.execute(db)
            similar_scores = dict(zip(similar_ids, similar_scores))
            if not self.query_linker:
                return similar_ids, similar_scores
            # TODO -- pre-select ids (this logic is wrong)
            query_linker = query_linker.select_using_ids(similar_ids)

        if not self.post_like:
            return query_linker.execute(db), similar_scores

        assert self.pre_like is None
        cursor = query_linker.select_ids.execute(db)
        query_ids = [str(document[self.primary_id]) for document in cursor]
        similar_ids, similar_scores = self.post_like.execute(db, ids=query_ids)
        similar_scores = dict(zip(similar_ids, similar_scores))

        post_query_linker = self.query_linker.select_using_ids(similar_ids)
        return post_query_linker.execute(db), similar_scores

    def execute(self, db):
        output, scores = self._execute(db)
        if isinstance(output, (pymongo.cursor.Cursor, mongomock.collection.Cursor)):
            return aladinCursor(
                raw_cursor=output,
                id_field='_id',
                scores=scores,
                encoders=db.encoders,
            )
        elif isinstance(output, dict):
            return Document(Document.decode(output, encoders=db.encoders))
        return output

    def download_update(self, db, id: str, key: str, bytes: bytearray) -> None:
        """
        Update to set the content of ``key`` in the document ``id``.

        :param db: The db to query
        :param id: The id to filter on
        :param key:
        :param bytes: The bytes to update
        """
        if self.collection is None:
            raise ValueError('collection cannot be None')
        update = {'$set': {f'{key}._content.bytes': bytes}}
        collection = db.databackend.get_table_or_collection(
            self.table_or_collection.identifier
        )
        return collection.update_one({'_id': id}, update)

    @property
    def select_table(self):
        return self.table_or_collection.find()


@dc.dataclass(repr=False)
class MongoQueryLinker(QueryLinker):
    @property
    def query_components(self):
        return self.table_or_collection.query_components

    @property
    def output_fields(self):
        out = {}
        for member in self.members:
            if hasattr(member, 'output_fields'):
                out.update(member.output_fields)
        return out

    def add_fold(self, fold):
        new_members = []
        for member in self.members:
            if hasattr(member, 'add_fold'):
                new_members.append(member.add_fold(fold))
            else:
                new_members.append(member)
        return MongoQueryLinker(
            table_or_collection=self.table_or_collection,
            members=new_members,
        )

    def outputs(self, **kwargs):
        new_members = []
        for member in self.members:
            if hasattr(member, 'outputs'):
                new_members.append(member.outputs(**kwargs))
            else:
                new_members.append(member)

        return MongoQueryLinker(
            table_or_collection=self.table_or_collection,
            members=new_members,
        )

    @property
    def select_ids(self):
        new_members = []
        for member in self.members:
            if hasattr(member, 'select_ids'):
                new_members.append(member.select_ids)
            else:
                new_members.append(member)

        return MongoQueryLinker(
            table_or_collection=self.table_or_collection,
            members=new_members,
        )

    def select_using_ids(self, ids):
        new_members = []
        for member in self.members:
            if hasattr(member, 'select_using_ids'):
                new_members.append(member.select_using_ids(ids))
            else:
                new_members.append(member)

        return MongoQueryLinker(
            table_or_collection=self.table_or_collection,
            members=new_members,
        )

    def _select_ids_of_missing_outputs(self, key: str, model: str, version: int):
        new_members = []
        for member in self.members:
            if hasattr(member, 'select_ids_of_missing_outputs'):
                new_members.append(
                    member.select_ids_of_missing_outputs(key, model, version=version)
                )
        return MongoQueryLinker(
            table_or_collection=self.table_or_collection,
            members=new_members,
        )

    def select_single_id(self, id):
        assert (
            len(self.members) == 1
            and self.members[0].type == QueryType.QUERY
            and hasattr(self.members[0].name, 'select_single_id')
        )
        return MongoQueryLinker(
            table_or_collection=self.table_or_collection,
            members=[self.members[0].select_single_id(id)],
        )

    def execute(self, db):
        parent = db.databackend.get_table_or_collection(
            self.table_or_collection.identifier
        )
        for member in self.members:
            parent = member.execute(parent)
        return parent


@dc.dataclass(repr=False)
class MongoInsert(Insert):
    one: bool = False

    def execute(self, db):
        collection = db.databackend.get_table_or_collection(
            self.table_or_collection.identifier
        )
        documents = [r.encode() for r in self.documents]
        insert_result = collection.insert_many(documents, **self.kwargs)
        return insert_result.inserted_ids

    @property
    def select_table(self):
        return self.table_or_collection.find()


@dc.dataclass(repr=False)
class MongoDelete(Delete):
    one: bool = False

    @property
    def collection(self):
        return self.table_or_collection

    def execute(self, db):
        collection = db.databackend.get_table_or_collection(
            self.table_or_collection.identifier
        )
        if self.one:
            return collection.delete_one(*self.args, **self.kwargs)
        delete_result = collection.delete_many(*self.args, **self.kwargs)
        return delete_result.deleted_ids


@dc.dataclass(repr=False)
class MongoUpdate(Update):
    update: Document
    filter: t.Dict
    one: bool = False
    args: t.Sequence = dc.field(default_factory=list)
    kwargs: t.Dict = dc.field(default_factory=dict)

    @property
    def select_table(self):
        return self.table_or_collection.find()

    def execute(self, db):
        collection = db.databackend.get_table_or_collection(
            self.table_or_collection.identifier
        )

        update = self.update
        if isinstance(self.update, Document):
            update = update.encode()

        if self.one:
            id = collection.find_one(self.filter, {'_id': 1})['_id']
            collection.update_one({'_id': id}, update, *self.args, **self.kwargs)
            return [id]

        ids = [r['_id'] for r in collection.find(self.filter, {'_id': 1})]
        collection.update_many({'_id': {'$in': ids}}, update, *self.args, **self.kwargs)
        return ids


@dc.dataclass(repr=False)
class MongoReplaceOne(Update):
    replacement: Document
    filter: t.Dict
    args: t.Sequence = dc.field(default_factory=list)
    kwargs: t.Dict = dc.field(default_factory=dict)

    @property
    def collection(self):
        return self.table_or_collection

    @property
    def select_table(self):
        return self.table_or_collection.find()

    def execute(self, db):
        collection = db.databackend.get_table_or_collection(
            self.table_or_collection.identifier
        )

        replacement = self.replacement
        if isinstance(replacement, Document):
            replacement = replacement.encode()

        id = collection.find_one(self.filter, {'_id': 1})['_id']
        collection.replace_one({'_id': id}, replacement, *self.args, **self.kwargs)
        return [id]


@dc.dataclass
class ChangeStream:
    """Request a stream of changes from a db

    :param collection: The collection to perform the query on
    :param args: Positional query arguments to ``pymongo.Collection.watch``
    :param kwargs: Named query arguments to ``pymongo.Collection.watch``
    """

    collection: 'Collection'
    args: t.Sequence = dc.field(default_factory=list)
    kwargs: t.Dict = dc.field(default_factory=dict)

    def __call__(self, db):
        collection = db.databackend.get_table_or_collection(self.collection)
        return collection.watch(**self.kwargs)


@dc.dataclass(repr=False)
class Collection(TableOrCollection):
    query_components: t.ClassVar[t.Dict] = {'find': Find, 'find_one': FindOne}
    type_id: t.ClassVar[str] = 'collection'

    primary_id: t.ClassVar[str] = '_id'

    def get_table(self, db):
        collection = db.databackend.get_table_or_collection(self.collection.identifier)
        return collection

    def change_stream(self, *args, **kwargs):
        return ChangeStream(
            collection=self.identifier,
            args=args,
            kwargs=kwargs,
        )

    def _get_query_linker(self, members) -> MongoQueryLinker:
        return MongoQueryLinker(members=members, table_or_collection=self)

    def _get_query(
        self,
        pre_like: t.Optional[Like] = None,
        query_linker: t.Optional[QueryLinker] = None,
        post_like: t.Optional[Like] = None,
        i: int = 0,
    ) -> MongoCompoundSelect:
        return MongoCompoundSelect(
            pre_like=pre_like,
            query_linker=query_linker,
            post_like=post_like,
            table_or_collection=self,
        )

    def _delete(self, *args, one: bool = False, **kwargs):
        return MongoDelete(args=args, kwargs=kwargs, table_or_collection=self, one=one)

    def _insert(self, documents, **kwargs):
        return MongoInsert(documents=documents, kwargs=kwargs, table_or_collection=self)

    def _update(self, filter, update, *args, one: bool = False, **kwargs):
        return MongoUpdate(
            filter=filter,
            update=update,
            args=args,
            kwargs=kwargs,
            table_or_collection=self,
            one=one,
        )

    def aggregate(
        self, *args, vector_index: t.Optional[str] = None, **kwargs
    ) -> Aggregate:
        return Aggregate(
            args=args,
            kwargs=kwargs,
            vector_index=vector_index,
            table_or_collection=self,
        )

    def delete_one(self, *args, **kwargs):
        return self._delete(*args, one=True, **kwargs)

    def replace_one(self, filter, replacement, *args, **kwargs):
        return MongoReplaceOne(
            filter=filter,
            replacement=replacement,
            args=args,
            kwargs=kwargs,
            table_or_collection=self,
        )

    def update_one(self, filter, update, *args, **kwargs):
        return self._update(filter, update, *args, one=True, **kwargs)

    def update_many(self, filter, update, *args, **kwargs):
        return self._update(filter, update, *args, one=False, **kwargs)

    def insert(self, *args, **kwargs):
        return self.insert_many(*args, **kwargs)

    def insert_many(self, *args, **kwargs):
        return self._insert(*args, **kwargs)

    def insert_one(self, document, *args, **kwargs):
        return self._insert([document], *args, **kwargs)

    def model_update(
        self,
        db,
        ids: t.List[t.Any],
        key: str,
        model: str,
        version: int,
        outputs: t.Sequence[t.Any],
        flatten: bool = False,
        **kwargs,
    ):
        document_embedded = kwargs.get('document_embedded', True)

        if key.startswith('_outputs'):
            key = key.split('.')[1]
        if not outputs:
            return
        if document_embedded:
            if flatten:
                raise AttributeError(
                    'Flattened outputs cannot be stored along with input documents.'
                    'Please use `document_embedded = False` option with flatten = True'
                )
            assert self.collection is not None
            collection = db.databackend.get_table_or_collection(self.identifier)
            collection.bulk_write(
                [
                    _UpdateOne(
                        {'_id': ObjectId(id)},
                        {'$set': {f'_outputs.{key}.{model}.{version}': outputs[i]}},
                    )
                    for i, id in enumerate(ids)
                ]
            )
        else:
            if flatten:
                bulk_docs = []
                for i, id in enumerate(ids):
                    _outputs = outputs[i]
                    if isinstance(_outputs, (list, tuple)):
                        for offset, output in enumerate(_outputs):
                            bulk_docs.append(
                                _InsertOne(
                                    {
                                        '_outputs': {
                                            key: {model: {str(version): output}}
                                        },
                                        '_source': ObjectId(id),
                                        '_offset': offset,
                                    }
                                )
                            )
                    else:
                        bulk_docs.append(
                            _InsertOne(
                                {
                                    '_outputs': {
                                        key: {model: {str(version): _outputs}}
                                    },
                                    '_source': ObjectId(id),
                                    '_offset': 0,
                                }
                            )
                        )

            else:
                bulk_docs = [
                    _InsertOne(
                        {
                            '_id': ObjectId(id),
                            '_outputs': {key: {model: {str(version): outputs[i]}}},
                        }
                    )
                    for i, id in enumerate(ids)
                ]

            collection_name = f'_outputs.{key}.{model}'
            collection = db.databackend.get_table_or_collection(collection_name)
            collection.bulk_write(bulk_docs)



"""
Functions from later standard libraries not available in Python 3.8
"""

from functools import lru_cache

__all__ = ('cache',)


# Implements functools.cache from Python 3.9
def cache(user_function, /):
    return lru_cache(maxsize=None)(user_function)



import click
import gridfs

from aladindb import logging
from aladindb.backends.base.artifact import ArtifactStore
from aladindb.misc.colors import Colors


class MongoArtifactStore(ArtifactStore):
    """
    Artifact store for MongoDB.

    :param conn: MongoDB client connection
    :param name: Name of database to host filesystem
    """

    def __init__(self, conn, name: str):
        super().__init__(name=name, conn=conn)
        self.db = self.conn[self.name]
        self.filesystem = gridfs.GridFS(self.db)

    def url(self):
        return self.conn.HOST + ':' + str(self.conn.PORT) + '/' + self.name

    def drop(self, force: bool = False):
        if not force:
            if not click.confirm(
                f'{Colors.RED}[!!!WARNING USE WITH CAUTION AS YOU '
                f'WILL LOSE ALL DATA!!!]{Colors.RESET} '
                'Are you sure you want to drop all artifacts? ',
                default=False,
            ):
                logging.warn('Aborting...')
        return self.db.client.drop_database(self.db.name)

    def delete(self, file_id: str):
        return self.filesystem.delete(file_id)

    def load_bytes(self, file_id: str):
        return self.filesystem.get(file_id).read()

    def save_artifact(self, serialized: bytes):
        return self.filesystem.put(serialized)

    def disconnect(self):
        """
        Disconnect the client
        """

        # TODO: implement me



import typing as t

import numpy
import torch

from aladindb.components.encoder import Encoder
from aladindb.ext.utils import str_shape


class EncodeTensor:
    def __init__(self, dtype):
        self.dtype = dtype

    def __call__(self, x):
        if x.dtype != self.dtype:
            raise TypeError(f'dtype was {x.dtype}, expected {self.dtype}')
        return memoryview(x.numpy()).tobytes()


class DecodeTensor:
    def __init__(self, dtype):
        self.dtype = torch.randn(1).type(dtype).numpy().dtype

    def __call__(self, bytes):
        array = numpy.frombuffer(bytes, dtype=self.dtype)
        return torch.from_numpy(array)


def tensor(dtype, shape: t.Sequence):
    """
    Create an encoder for a tensor of a given dtype and shape.

    :param dtype: The dtype of the tensor.
    :param shape: The shape of the tensor.
    """
    return Encoder(
        identifier=f'{str(dtype)}[{str_shape(shape)}]',
        encoder=EncodeTensor(dtype),
        decoder=DecodeTensor(dtype),
        shape=shape,
    )



import base64
import json
from functools import lru_cache

import requests

from aladindb import CFG, logging
from aladindb.base import exceptions
from aladindb.ext.utils import aladinencode


@lru_cache(maxsize=None)
def _handshake(service: str):
    endpoint = 'handshake/config'
    cfg = json.dumps(CFG.comparables)
    try:
        _request_server(service, args={'cfg': cfg}, endpoint=endpoint)
    except Exception:
        raise Exception("Incompatible configuration")


def _request_server(
    service: str = 'vector_search', data=None, endpoint='add', args={}, type='post'
):
    service_uri = 'http://' + ''.join(getattr(CFG.cluster, service).split('://')[1:])

    url = service_uri + '/' + endpoint
    logging.debug(f'Trying to connect {service} at {url} method: {type}')

    if type == 'post':
        data = aladinencode(data)
        if isinstance(data, dict):
            if '_content' in data:
                data['_content']['bytes'] = base64.b64encode(
                    data['_content']['bytes']
                ).decode()
        response = requests.post(url, json=data, params=args)
        result = json.loads(response.content)
    else:
        response = requests.get(url, params=args)
        result = None
    if response.status_code != 200:
        error = json.loads(response.content)
        msg = f'Server error at {service} with {response.status_code} :: {error}'
        raise exceptions.ServiceRequestException(msg)
    return result


def request_server(
    service: str = 'vector_search', data=None, endpoint='add', args={}, type='post'
):
    _handshake(service)
    return _request_server(
        service=service, data=data, endpoint=endpoint, args=args, type=type
    )



import typing as t
import uuid

from aladindb import logging
from aladindb.backends.base.compute import ComputeBackend


class LocalComputeBackend(ComputeBackend):
    """
    A mockup backend for running jobs locally.
    """

    def __init__(
        self,
    ):
        self.__outputs: t.Dict = {}

    @property
    def type(self) -> str:
        return "local"

    @property
    def name(self) -> str:
        return "local"

    def submit(self, function: t.Callable, *args, **kwargs) -> str:
        """
        Submits a function for local execution.

        :param function: The function to be executed.
        """
        logging.info(f"Submitting job. function:{function}")
        future = function(*args, **kwargs)

        future_key = str(uuid.uuid4())
        self.__outputs[future_key] = future

        logging.success(f"Job submitted.  function:{function} future:{future_key}")
        return future_key

    @property
    def tasks(self) -> t.Dict[str, t.Any]:
        """
        List for all pending tasks
        """
        return self.__outputs

    def wait_all(self) -> None:
        """
        Waits for all pending tasks to complete.
        """
        pass

    def result(self, identifier: str) -> t.Any:
        """
        Retrieves the result of a previously submitted task.
        Note: This will block until the future is completed.

        :param identifier: The identifier of the submitted task.
        """
        return self.__outputs[identifier]

    def disconnect(self) -> None:
        """
        Disconnect the local client.
        """
        pass

    def shutdown(self) -> None:
        """
        Shuts down the local cluster.
        """
        pass



import sys

import click

from aladindb.cli import app, config, info
from aladindb.cli.serve import cdc, local_cluster, vector_search

__all__ = 'config', 'info', 'local_cluster', 'vector_search', 'cdc'


def run():
    """
    Entrypoint for the CLI. This is the function that is called when the
    user runs `python -m aladindb`.
    """
    try:
        app(standalone_mode=False)
    except click.ClickException as e:
        return f'{e.__class__.__name__}: {e.message}'
    except click.Abort:
        return 'Aborted'


if __name__ == '__main__':
    sys.exit(run())



from aladindb import CFG
from aladindb.base.datalayer import Datalayer
from aladindb.components.listener import Listener
from aladindb.server import app as aladinapp

assert isinstance(CFG.cluster.cdc, str), "cluster.cdc should be set with a valid uri"
port = int(CFG.cluster.cdc.split(':')[-1])
app = aladinapp.aladinApp('cdc', port=port)


@app.startup
def cdc_startup(db: Datalayer):
    db.cdc.start()


@app.add('/listener/add', method='get')
def add_listener(name: str, db: Datalayer = aladinapp.DatalayerDependency()):
    """
    Endpoint for adding a listener to cdc
    """
    listener = db.load('listener', name)
    assert isinstance(listener, Listener)
    db.cdc.add(listener)


@app.add('/listener/delete', method='get')
def remove_listener(name: str, db: Datalayer = aladinapp.DatalayerDependency()):
    """
    Endpoint for removing a listener from cdc
    """
    listener = db.load('listener', name)
    assert isinstance(listener, Listener)
    on = listener.select.table_or_collection.identifier
    db.cdc.stop(on)  # type: ignore[arg-type]



import dataclasses as dc
import enum
import typing as t
from abc import ABC, abstractmethod, abstractproperty
from typing import Any

from aladindb import logging
from aladindb.base.document import Document
from aladindb.base.serializable import Serializable, Variable

GREEN = '\033[92m'
BOLD = '\033[1m'
END = '\033[0m}'


class _ReprMixin(ABC):
    @abstractmethod
    def repr_(self) -> str:
        pass

    def __repr__(self) -> str:
        return (
            f'<{self.__class__.__module__}.{self.__class__.__name__}'
            f'[\n    {GREEN + BOLD}{self.repr_()}{END}\n] object at {hex(id(self))}>'
        )


@dc.dataclass(repr=False)
class Select(Serializable, ABC):
    """
    Base class for all select queries.
    """

    @abstractproperty
    def id_field(self):
        pass

    @property
    def query_components(self):
        return self.table_or_collection.query_components

    def model_update(
        self,
        db,
        ids: t.List[str],
        key: str,
        model: str,
        version: int,
        outputs: t.Sequence[t.Any],
        **kwargs,
    ):
        """
        Update model outputs for a set of ids.

        :param db: The DB instance to use
        :param ids: The ids to update
        :param key: The key to update
        :param model: The model to update
        :param outputs: The outputs to update
        """
        return self.table_or_collection.model_update(
            db=db,
            ids=ids,
            key=key,
            model=model,
            outputs=outputs,
            version=version,
            **kwargs,
        )

    @abstractproperty
    def select_table(self):
        pass

    @abstractmethod
    def add_fold(self, fold: str) -> 'Select':
        pass

    @abstractmethod
    def select_using_ids(self, ids: t.Sequence[str]) -> 'Select':
        pass

    @abstractproperty
    def select_ids(self) -> 'Select':
        pass

    @abstractmethod
    def select_ids_of_missing_outputs(
        self, key: str, model: str, version: int
    ) -> 'Select':
        pass

    @abstractmethod
    def select_single_id(self, id: str) -> 'Select':
        pass

    @abstractmethod
    def execute(self, db):
        """
        Execute the query on the DB instance.
        """
        pass


@dc.dataclass(repr=False)
class Delete(Serializable, ABC):
    """
    Base class for all deletion queries

    :param table_or_collection: The table or collection that this query is linked to
    """

    table_or_collection: 'TableOrCollection'
    args: t.Sequence = dc.field(default_factory=list)
    kwargs: t.Dict = dc.field(default_factory=dict)

    @abstractmethod
    def execute(self, db):
        pass


@dc.dataclass(repr=False)
class Update(Serializable, ABC):
    """
    Base class for all update queries

    :param table_or_collection: The table or collection that this query is linked to
    """

    table_or_collection: 'TableOrCollection'

    @abstractmethod
    def select_table(self):
        pass

    @abstractmethod
    def execute(self, db):
        pass


@dc.dataclass(repr=False)
class CompoundSelect(_ReprMixin, Select, ABC):
    """
    A query with multiple parts.

    like ----> select ----> like

    :param table_or_collection: The table or collection that this query is linked to
    :param pre_like: The pre_like part of the query (e.g. ``table.like(...)...``)
    :param post_like: The post_like part of the query
                      (e.g. ``table.filter(...)....like(...)``)
    :param query_linker: The query linker that is responsible for linking the
                         query chain. E.g. ``table.filter(...).select(...)``.
    :param i: The index of the query in the query chain
    """

    table_or_collection: 'TableOrCollection'
    pre_like: t.Optional['Like'] = None
    post_like: t.Optional['Like'] = None
    query_linker: t.Optional['QueryLinker'] = None

    @abstractproperty
    def output_fields(self):
        pass

    @property
    def id_field(self):
        return self.primary_id

    @property
    def primary_id(self):
        return self.table_or_collection.primary_id

    def add_fold(self, fold: str):
        assert self.pre_like is None
        assert self.post_like is None
        assert self.query_linker is not None
        return self._query_from_parts(
            table_or_collection=self.table_or_collection,
            query_linker=self.query_linker.add_fold(fold),
        )

    @property
    def select_ids(self):
        """
        Query which selects the same documents/ rows but only ids.
        """

        assert self.pre_like is None
        assert self.post_like is None

        return self._query_from_parts(
            table_or_collection=self.table_or_collection,
            query_linker=self.query_linker.select_ids,
        )

    def select_ids_of_missing_outputs(self, key: str, model: str, version: int):
        """
        Query which selects ids where outputs are missing.
        """

        assert self.pre_like is None
        assert self.post_like is None
        assert self.query_linker is not None

        return self._query_from_parts(
            table_or_collection=self.table_or_collection,
            query_linker=self.query_linker._select_ids_of_missing_outputs(
                key=key, model=model, version=version
            ),
        )

    def select_single_id(self, id: str):
        """
        Query which selects a single id.

        :param id: The id to select.
        """
        assert self.pre_like is None
        assert self.post_like is None
        assert self.query_linker is not None

        return self._query_from_parts(
            table_or_collection=self.table_or_collection,
            query_linker=self.query_linker.select_single_id(id),
        )

    def select_using_ids(self, ids):
        """
        Subset a query to only these ids.

        :param ids: The ids to subset to.
        """

        assert self.pre_like is None
        assert self.post_like is None

        return self._query_from_parts(
            table_or_collection=self.table_or_collection,
            query_linker=self.query_linker.select_using_ids(ids),
        )

    def repr_(self):
        """
        String representation of the query.
        """

        components = []
        components.append(str(self.table_or_collection.identifier))
        if self.pre_like:
            components.append(str(self.pre_like))
        if self.query_linker:
            components.extend(self.query_linker.repr_().split('.')[1:])
        if self.post_like:
            components.append(str(self.post_like))
        return '.'.join(components)

    @classmethod
    def _query_from_parts(
        cls,
        table_or_collection,
        pre_like=None,
        post_like=None,
        query_linker=None,
    ):
        return cls(
            table_or_collection=table_or_collection,
            pre_like=pre_like,
            post_like=post_like,
            query_linker=query_linker,
        )

    def _get_query_component(
        self,
        name: str,
        type: str,
        args: t.Optional[t.Sequence] = None,
        kwargs: t.Optional[t.Dict] = None,
    ):
        query_component_cls = self.table_or_collection.query_components.get(
            name, QueryComponent
        )
        return query_component_cls(name, type=type, args=args, kwargs=kwargs)

    @abstractmethod
    def _get_query_linker(cls, table_or_collection, members) -> 'QueryLinker':
        pass

    def __getattr__(self, name):
        assert self.post_like is None
        if self.query_linker is not None:
            query_linker = getattr(self.query_linker, name)
        else:
            query_linker = self._get_query_linker(
                self.table_or_collection,
                members=[self._get_query_component(name, type=QueryType.ATTR)],
            )
        return self._query_from_parts(
            table_or_collection=self.table_or_collection,
            pre_like=self.pre_like,
            query_linker=query_linker,
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        assert self.post_like is None
        assert self.query_linker is not None
        return self._query_from_parts(
            table_or_collection=self.table_or_collection,
            pre_like=self.pre_like,
            query_linker=self.query_linker(*args, **kwargs),
        )

    @abstractmethod
    def execute(self, db):
        """
        Execute the compound query on the DB instance.

        :param db: The DB instance to use
        """

    def like(self, r: Document, vector_index: str, n: int = 10):
        assert self.query_linker is not None
        assert self.pre_like is None
        return self._query_from_parts(
            table_or_collection=self.table_or_collection,
            pre_like=None,
            query_linker=self.query_linker,
            post_like=Like(r=r, n=n, vector_index=vector_index),
        )


@dc.dataclass(repr=False)
class Insert(_ReprMixin, Serializable, ABC):
    """
    Base class for all insert queries.

    :param table_or_collection: The table or collection that this query is linked to
    :param documents: The documents to insert
    :param refresh: Whether to refresh the task-graph after inserting
    :param verbose: Whether to print the progress of the insert
    :param kwargs: Any additional keyword arguments to pass to the insert method
    :param encoders: The encoders to use to encode the documents
    """

    table_or_collection: 'TableOrCollection'
    documents: t.Sequence['Document'] = dc.field(default_factory=list)
    verbose: bool = True
    kwargs: t.Dict = dc.field(default_factory=dict)

    def repr_(self):
        documents_str = (
            str(self.documents)[:25] + '...'
            if len(self.documents) > 25
            else str(self.documents)
        )
        return f'{self.table_or_collection.identifier}.insert_many({documents_str})'

    @abstractmethod
    def select_table(self):
        pass

    @abstractmethod
    def execute(self, parent: t.Any):
        """
        Insert the data.

        :param parent: The parent instance to use for insertion
        """
        pass

    def to_select(self, ids=None):
        if ids is None:
            ids = [r['_id'] for r in self.documents]
        return self.table.find({'_id': ids})


class QueryType(str, enum.Enum):
    """
    The type of a query. Either `query` or `attr`.
    """

    QUERY = 'query'
    ATTR = 'attr'


@dc.dataclass(repr=False)
class QueryComponent(Serializable):
    """
    This is a representation of a single query object in ibis query chain.
    This is used to build a query chain that can be executed on a database.
    Query will be executed in the order they are added to the chain.

    If we have a query chain like this:
        query = t.select(['id', 'name']).limit(10)
    here we have 2 query objects, `select` and `limit`.

    `select` will be wrapped with this class and added to the chain.

    :param name: The name of the query
    :param type: The type of the query, either `query` or `attr`
    :param args: The arguments to pass to the query
    :param kwargs: The keyword arguments to pass to the query
    """

    name: str
    type: str = QueryType.ATTR
    args: t.Sequence = dc.field(default_factory=list)
    kwargs: t.Dict = dc.field(default_factory=dict)

    def repr_(self) -> str:
        if self.type == QueryType.ATTR:
            return self.name

        def to_str(x):
            if isinstance(x, str):
                return f"'{x}'"
            elif isinstance(x, list):
                return f'[{", ".join([to_str(a) for a in x])}]'
            elif isinstance(x, dict):
                return str({k: to_str(v) for k, v in x.items()})
            elif isinstance(x, _ReprMixin):
                return x.repr_()
            else:
                return str(x)

        args_as_strs = [to_str(a) for a in self.args]
        args_as_strs += [f'{k}={to_str(v)}' for k, v in self.kwargs.items()]
        joined = ', '.join(args_as_strs)
        return f'{self.name}({joined})'

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        try:
            assert (
                self.type == QueryType.ATTR
            ), '__call__ method must be called on an attribute query'
        except AssertionError as e:
            logging.warn('QUERY_COMPONENT: ' + self.name)
            raise e
        return type(self)(
            name=self.name,
            type=QueryType.QUERY,
            args=args,
            kwargs=kwargs,
        )

    def execute(self, parent: t.Any):
        if self.type == QueryType.ATTR:
            return getattr(parent, self.name)
        assert self.type == QueryType.QUERY
        parent = getattr(parent, self.name)(*self.args, **self.kwargs)
        return parent


@dc.dataclass(repr=False)
class QueryLinker(_ReprMixin, Serializable, ABC):
    """
    This class is responsible for linking together a query using
    `getattr` and `__call__`.

    This allows ``aladindb`` to serialize queries from a range of APIs.
    Intuitively this allows us to do something like this:

    >>> collection.find({}).limit(10).sort('name')
    -->
    [
        ('<NAME>', <ARGS>, <KWARGS>),
        ('find', {}, None),
        ('limit', 10, None),
        ('sort', 'name', None),
    ]

    table.filter(t.select('id') == '1')

    :param table_or_collection: The table or collection that this query is linked to.
    :param members: The members of the query chain.
    """

    table_or_collection: 'TableOrCollection'
    members: t.List = dc.field(default_factory=list)

    @property
    def query_components(self):
        return self.table_or_collection.query_components

    def repr_(self) -> str:
        return (
            f'{self.table_or_collection.identifier}'
            + '.'
            + '.'.join([m.repr_() for m in self.members])
        )

    def _get_query_component(self, k):
        if k in self.query_components:
            return self.query_components[k](name=k, type=QueryType.ATTR)
        return QueryComponent(name=k, type=QueryType.ATTR)

    @classmethod
    def _get_query_linker(cls, table_or_collection, members):
        return cls(
            table_or_collection=table_or_collection,
            members=members,
        )

    def __getattr__(self, k):
        return self._get_query_linker(
            self.table_or_collection,
            members=[*self.members, self._get_query_component(k)],
        )

    @property
    @abstractmethod
    def select_ids(self):
        pass

    @abstractmethod
    def select_single_id(self, id):
        pass

    @abstractmethod
    def select_using_ids(self, ids):
        pass

    def __call__(self, *args, **kwargs):
        members = [*self.members[:-1], self.members[-1](*args, **kwargs)]
        return type(self)(table_or_collection=self.table_or_collection, members=members)

    @abstractmethod
    def execute(self, db):
        pass


@dc.dataclass
class Like(Serializable):
    """
    Base class for all like (vector-search) queries.

    :param r: The item to be converted to a vector, to search with.
    :param vector_index: The vector index to use
    :param n: The number of results to return
    """

    r: t.Union[t.Dict, Document]
    vector_index: str
    n: int = 10

    def execute(self, db, ids: t.Optional[t.Sequence[str]] = None):
        return db.select_nearest(
            like=self.r,
            vector_index=self.vector_index,
            ids=ids,
            n=self.n,
        )


@dc.dataclass
class TableOrCollection(Serializable, ABC):
    """
    This is a representation of an SQL table in ibis.

    :param identifier: The name of the table
    """

    query_components: t.ClassVar[t.Dict] = {}
    type_id: t.ClassVar[str] = 'table_or_collection'
    identifier: t.Union[str, Variable]

    @abstractmethod
    def _get_query_linker(self, members) -> QueryLinker:
        pass

    def _get_query_component(self, name: str) -> QueryComponent:
        return self.query_components.get(name, QueryComponent)(
            name=name, type=QueryType.ATTR
        )

    @abstractmethod
    def model_update(
        self,
        db,
        ids: t.List[t.Any],
        key: str,
        model: str,
        version: int,
        outputs: t.Sequence[t.Any],
        flatten: bool = False,
        **kwargs,
    ):
        pass

    @abstractmethod
    def insert(self, documents: t.Sequence[Document], **kwargs) -> Insert:
        pass

    @abstractmethod
    def _get_query(
        self,
        pre_like: t.Optional[Like] = None,
        query_linker: t.Optional[QueryLinker] = None,
        post_like: t.Optional[Like] = None,
    ) -> CompoundSelect:
        pass

    def __getattr__(self, k: str) -> 'CompoundSelect':
        # This method is responsible for dynamically creating a query chain,
        # which can be executed on a database. This is done by creating a
        # QueryLinker object, which is a representation of a query chain.
        # Under the hood, this is done by creating a QueryChain object, which
        # is a representation of a query chain.
        query_linker = self._get_query_linker([self._get_query_component(k)])
        return self._get_query(query_linker=query_linker)

    def like(
        self,
        r: Document,
        vector_index: str,
        n: int = 10,
    ):
        """
        This method appends a query to the query chain where the query is repsonsible
        for performing a vector search on the parent query chain inputs.

        :param r: The vector to search for
        :param vector_index: The vector index to use
        :param n: The number of results to return
        """
        return self._get_query(
            pre_like=Like(
                r=r,
                n=n,
                vector_index=vector_index,
            ),
        )

    @abstractmethod
    def _insert(
        self,
        documents: t.Sequence[Document],
        *,
        refresh: bool = False,
        encoders: t.Sequence = (),
        verbose: bool = True,
        **kwargs,
    ):
        pass


@dc.dataclass
class RawQuery:
    query: t.Any

    @abstractmethod
    def execute(self, db):
        '''
        A raw query method which executes the query and returns the result
        '''



import datetime
import json
import os
import platform
import socket
import sys
from pathlib import Path

from aladindb import ROOT

from . import command

PYPROJECT = ROOT / 'pyproject.toml'


@command(help='Print information about the current machine and installation')
def info():
    print('```')
    print(json.dumps(_get_info(), default=str, indent=2))
    print('```')


def _get_info():
    return {
        'cfg': _cfg(),
        'cwd': Path.cwd(),
        'freeze': _freeze(),
        'hostname': socket.gethostname(),
        'os_uname': list(os.uname()),
        'package_versions': _package_versions(),
        'platform': _platform(),
        'startup_time': datetime.datetime.now(),
        'aladin_db_root': ROOT,
        'sys': _sys(),
    }


def _cfg():
    try:
        from aladindb import CFG

        return CFG.dict()
    except Exception:
        return '(CFG not yet commited)'


def _freeze():
    try:
        from pip._internal.operations.freeze import freeze

        return list(freeze())
    except Exception as e:
        return [f'Freeze failed with {e}']


def _package_versions():
    return {}


def _platform():
    return {
        'platform': platform.platform(),
        'python_version': platform.python_version(),
    }


def _sys():
    return {'argv': sys.argv, 'path': sys.path}



import typing as t

import numpy as np

from aladindb import CFG
from aladindb.misc.server import request_server
from aladindb.vector_search.base import BaseVectorSearcher, VectorItem

if t.TYPE_CHECKING:
    from aladindb.base.datalayer import Datalayer


class FastVectorSearcher(BaseVectorSearcher):
    def __init__(self, db: 'Datalayer', vector_searcher, vector_index: str):
        self.searcher = vector_searcher
        self.vector_index = vector_index

        if CFG.cluster.is_remote_vector_search:
            if not db.server_mode:
                request_server(
                    service='vector_search',
                    endpoint='create/search',
                    args={
                        'vector_index': self.vector_index,
                    },
                    type='get',
                )

    def __len__(self):
        return len(self.searcher)

    def add(self, items: t.Sequence[VectorItem]) -> None:
        """
        Add items to the index.

        :param items: t.Sequence of VectorItems
        """
        vector_items = [{'vector': i.vector, 'id': i.id} for i in items]
        if CFG.cluster.is_remote_vector_search:
            request_server(
                service='vector_search',
                data=vector_items,
                endpoint='add/search',
                args={
                    'vector_index': self.vector_index,
                },
            )
            return

        return self.searcher.add(items)

    def delete(self, ids: t.Sequence[str]) -> None:
        """
        Remove items from the index

        :param ids: t.Sequence of ids of vectors.
        """
        if CFG.cluster.is_remote_vector_search:
            request_server(
                service='vector_search',
                data=ids,
                endpoint='delete/search',
                args={
                    'vector_index': self.vector_index,
                },
            )
            return

        return self.searcher.delete(ids)

    def find_nearest_from_id(
        self,
        _id,
        n: int = 100,
        within_ids: t.Sequence[str] = (),
    ) -> t.Tuple[t.List[str], t.List[float]]:
        """
        Find the nearest vectors to the vector with the given id.

        :param _id: id of the vector
        :param n: number of nearest vectors to return
        """
        if CFG.cluster.is_remote_vector_search:
            response = request_server(
                service='vector_search',
                endpoint='query/id/search',
                args={'vector_index': self.vector_index, 'n': n, 'id': _id},
            )
            return response['ids'], response['scores']

        return self.searcher.find_nearest_from_id(_id, n=n, within_ids=within_ids)

    def find_nearest_from_array(
        self,
        h: np.typing.ArrayLike,
        n: int = 100,
        within_ids: t.Sequence[str] = (),
    ) -> t.Tuple[t.List[str], t.List[float]]:
        """
        Find the nearest vectors to the given vector.

        :param h: vector
        :param n: number of nearest vectors to return
        """
        if CFG.cluster.is_remote_vector_search:
            response = request_server(
                service='vector_search',
                data=h,
                endpoint='query/search',
                args={'vector_index': self.vector_index, 'n': n},
            )
            return response['ids'], response['scores']

        return self.searcher.find_nearest_from_array(h=h, n=n, within_ids=within_ids)



import contextlib
import traceback
import typing as t

if t.TYPE_CHECKING:
    from aladindb.base.datalayer import Datalayer


def method_job(
    cfg,
    type_id,
    identifier,
    method_name,
    args,
    kwargs,
    job_id,
    dependencies=(),
    local=True,
    db: t.Optional['Datalayer'] = None,
):
    """
    Run a method on a component in the database.

    :param cfg: user config
    :param type_id: type of component
    :param identifier: identifier of component
    :param method_name: name of method to run
    :param args: positional arguments to pass to the method
    :param kwargs: keyword arguments to pass to the method
    :param job_id: unique identifier for this job
    :param dependencies: other jobs that this job depends on
    """
    from aladindb.base.build import build_datalayer
    from aladindb.base.configs import build_config

    if db is None:
        cfg = build_config(cfg)
        cfg.force_set('cluster.compute', None)
        db = build_datalayer(cfg)

    component = db.load(type_id, identifier)
    method = getattr(component, method_name)
    db.metadata.update_job(job_id, 'status', 'running')

    try:
        if local:
            method(*args, db=db, **kwargs)
        else:
            handle_function_output(
                method,
                db=db,
                job_id=job_id,
                args=args,
                kwargs=kwargs,
            )
    except Exception as e:
        tb = traceback.format_exc()
        db.metadata.update_job(job_id, 'status', 'failed')
        db.metadata.update_job(job_id, 'msg', tb)
        raise e
    db.metadata.update_job(job_id, 'status', 'success')


class Logger:
    def __init__(self, database, id_, stream='stdout'):
        self.database = database
        self.id_ = id_
        self.stream = stream

    def write(self, message):
        self.database.metadata.write_output_to_job(
            self.id_, message, stream=self.stream
        )

    def flush(self):
        pass


def handle_function_output(function, db, job_id, args, kwargs):
    with contextlib.redirect_stdout(Logger(db, job_id)):
        with contextlib.redirect_stderr(Logger(db, job_id, stream='stderr')):
            return function(db=db, *args, **kwargs)


def callable_job(
    cfg,
    function_to_call,
    args,
    kwargs,
    job_id,
    dependencies=(),
    local=True,
    db: t.Optional['Datalayer'] = None,
):
    from aladindb.base.build import build_datalayer
    from aladindb.base.configs import build_config

    if db is None:
        cfg = build_config(cfg)
        cfg.force_set('cluster.compute', None)
        db = build_datalayer(cfg)
    db.metadata.update_job(job_id, 'status', 'running')
    output = None
    try:
        if local:
            output = function_to_call(*args, db=db, **kwargs)
        else:
            output = handle_function_output(
                function_to_call,
                db=db,
                job_id=job_id,
                args=args,
                kwargs=kwargs,
            )
    except Exception as e:
        tb = traceback.format_exc()
        db.metadata.update_job(job_id, 'status', 'failed')
        db.metadata.update_job(job_id, 'msg', tb)
        raise e
    else:
        db.metadata.update_job(job_id, 'status', 'success')
    return output



import dataclasses as dc

import pandas

from aladindb.base.cursor import aladinCursor
from aladindb.base.document import Document


@dc.dataclass
class aladinIbisResult(aladinCursor):
    def as_pandas(self):
        return pandas.DataFrame([Document(r).unpack() for r in self.raw_cursor])

    def __getitem__(self, item):
        return self.raw_cursor[item]

    def __len__(self):
        return len(self.raw_cursor)



import dataclasses as dc
import typing as t
from functools import cached_property

from aladindb.backends.ibis.field_types import dtype
from aladindb.base.configs import CFG
from aladindb.components.component import Component
from aladindb.components.encoder import Encoder
from aladindb.misc.annotations import public_api


@public_api(stability='beta')
@dc.dataclass(kw_only=True)
class Schema(Component):
    """
    A component carrying the information
    about the types or `Encoders` of a `Table`
    {component_parameters}
    :param fields: A mapping of field names to types or `Encoders`
    """

    __doc__ = __doc__.format(component_parameters=Component.__doc__)

    type_id: t.ClassVar[str] = 'schema'

    fields: t.Mapping[str, t.Union[Encoder, str]]

    def __post_init__(self):
        super().__post_init__()

        assert self.identifier is not None, 'Schema must have an identifier'
        assert self.fields is not None, 'Schema must have fields'
        self.fields['_fold'] = dtype('str')

    def pre_create(self, db) -> None:
        for v in self.fields.values():
            if isinstance(v, Encoder):
                db.add(v)
        return super().pre_create(db)

    @property
    def raw(self):
        return {
            k: (v.identifier if not isinstance(v, Encoder) else CFG.bytes_encoding)
            for k, v in self.fields.items()
        }

    @cached_property
    def encoded_types(self):
        return [k for k, v in self.fields.items() if isinstance(v, Encoder)]

    @cached_property
    def trivial(self):
        return not any([isinstance(v, Encoder) for v in self.fields.values()])

    @property
    def encoders(self):
        for v in self.fields.values():
            if isinstance(v, Encoder):
                yield v

    def decode(self, data: t.Mapping[str, t.Any]) -> t.Mapping[str, t.Any]:
        """
        Decode data using the schema's encoders

        :param data: data to decode
        """

        if self.trivial:
            return data
        decoded = {}
        for k, v in data.items():
            if k in self.encoded_types:
                field = self.fields[k]
                assert isinstance(field, Encoder)
                v = field.decode(v)
            decoded[k] = v
        return decoded

    def encode(self, data: t.Mapping[str, t.Any]):
        """
        Encode data using the schema's encoders

        :param data: data to encode
        """
        if self.trivial:
            return data
        encoded_data = {}
        for k, v in data.items():
            if k in self.fields and isinstance(self.fields[k], Encoder):
                field_encoder = self.fields[k]
                assert callable(field_encoder)
                encoded_data.update({k: field_encoder(v).encode()})
            else:
                encoded_data.update({k: v})
        return encoded_data



import os
import socket
import uuid
from sys import stderr

from loguru import logger
from loki_logger_handler.loki_logger_handler import LoguruFormatter, LokiLoggerHandler
from tqdm import tqdm

from aladindb.base.config import LogLevel, LogType

from .configs import CFG

__all__ = ('Logging',)


class Logging:
    if CFG.logging_type == LogType.LOKI:  # Send logs to Loki
        custom_handler = LokiLoggerHandler(
            url=os.environ["LOKI_URI"],
            labels={"application": "Test", "environment": "Develop"},
            labelKeys={},
            timeout=10,
            defaultFormatter=LoguruFormatter(),
        )

        logger.configure(handlers=[{"sink": custom_handler, "serialize": True}])
    else:
        # Replace default logger with a custom aladin format.
        logger.remove()

        # Enrich logger with additional information.
        logger.configure(
            extra={
                "hostname": socket.gethostname(),
                "session_id": str(uuid.uuid4()),
            }
        )

        fmt = (
            "<green> {time:YYYY-MMM-DD HH:mm:ss.SS}</green>"
            "| <level>{level: <8}</level> "
            "| <cyan>{extra[hostname]: <8}</cyan>"
            "| <cyan>{extra[session_id]}</cyan>"
            "| <cyan>{name}</cyan>:<cyan>{line: <4}</cyan> "
            "| <level>{message}</level>"
        )

        # DEBUG until WARNING are sent to stdout.
        logger.add(
            lambda msg: tqdm.write(msg, end=""),
            format=fmt,
            level=CFG.log_level,
            filter=lambda record: record["level"].no < 40,
            colorize=True,
        )

        # ERROR and above sent to stderr
        # https://loguru.readthedocs.io/en/stable/api/logger.html
        logger.add(
            stderr,
            format=fmt,
            level=LogLevel.ERROR,
            colorize=True,
        )

    # Set Multi-Key loggers
    # Example: logging.info("param 1", "param 2", ..)
    @staticmethod
    def multikey_debug(msg: str, *args):
        logger.opt(depth=1).debug(" ".join(map(str, (msg, *args))))

    @staticmethod
    def multikey_info(msg: str, *args):
        logger.opt(depth=1).info(" ".join(map(str, (msg, *args))))

    @staticmethod
    def multikey_success(msg: str, *args):
        logger.opt(depth=1).success(" ".join(map(str, (msg, *args))))

    @staticmethod
    def multikey_warn(msg: str, *args):
        logger.opt(depth=1).warning(" ".join(map(str, (msg, *args))))

    @staticmethod
    def multikey_error(msg: str, *args):
        logger.opt(depth=1).error(" ".join(map(str, (msg, *args))))

    @staticmethod
    def multikey_exception(msg: str, *args, e=None):
        logger.opt(depth=1, exception=e).error(" ".join(map(str, (msg, *args))))

    debug = multikey_debug
    info = multikey_info
    success = multikey_success
    warn = multikey_warn
    error = multikey_error
    exception = multikey_exception



import os
import typing as t

import click
import pymongo

from aladindb import logging
from aladindb.backends.base.data_backend import BaseDataBackend
from aladindb.backends.mongodb.artifacts import MongoArtifactStore
from aladindb.backends.mongodb.metadata import MongoMetaDataStore
from aladindb.base.serializable import Serializable
from aladindb.misc.colors import Colors
from aladindb.misc.special_dicts import MongoStyleDict


class MongoDataBackend(BaseDataBackend):
    """
    Data backend for MongoDB.

    :param conn: MongoDB client connection
    :param name: Name of database to host filesystem
    """

    id_field = '_id'

    def __init__(self, conn: pymongo.MongoClient, name: str):
        super().__init__(conn=conn, name=name)
        self._db = self.conn[self.name]

    def url(self):
        return self.conn.HOST + ':' + str(self.conn.PORT) + '/' + self.name

    @property
    def db(self):
        return self._db

    def build_metadata(self):
        return MongoMetaDataStore(self.conn, self.name)

    def build_artifact_store(self):
        from mongomock import MongoClient as MockClient

        if isinstance(self.conn, MockClient):
            from aladindb.backends.local.artifacts import (
                FileSystemArtifactStore,
            )

            os.makedirs(f'/tmp/{self.name}', exist_ok=True)
            return FileSystemArtifactStore(f'/tmp/{self.name}')
        return MongoArtifactStore(self.conn, f'_filesystem:{self.name}')

    def drop(self, force: bool = False):
        if not force:
            if not click.confirm(
                f'{Colors.RED}[!!!WARNING USE WITH CAUTION AS YOU '
                f'WILL LOSE ALL DATA!!!]{Colors.RESET} '
                'Are you sure you want to drop the data-backend? ',
                default=False,
            ):
                logging.warn('Aborting...')
        return self.db.client.drop_database(self.db.name)

    def get_table_or_collection(self, identifier):
        return self._db[identifier]

    def set_content_bytes(self, r, key, bytes_):
        if not isinstance(r, MongoStyleDict):
            r = MongoStyleDict(r)
        r[f'{key}._content.bytes'] = bytes_
        return r

    def unset_outputs(self, info: t.Dict):
        select = Serializable.deserialize(info['select'])
        logging.info(f'unsetting output field _outputs.{info["key"]}.{info["model"]}')
        doc = {'$unset': {f'_outputs.{info["key"]}.{info["model"]}': 1}}
        update = select.update(doc)
        return self.db[select.collection].update_many(update.filter, update.update)

    def list_vector_indexes(self):
        indexes = []
        for coll in self.db.list_collection_names():
            i = self.db.command({'listSearchIndexes': coll})
            try:
                batch = i['cursor']['firstBatch'][0]
            except IndexError:
                continue
            if '_outputs' in batch['latestDefinition']['mappings']['fields']:
                indexes.append(batch['name'])
        return indexes

    def delete_vector_index(self, vector_index):
        """
        Delete a vector index in the data backend if an Atlas deployment.

        :param vector_index: vector index to delete
        """
        # see `VectorIndex` class for details
        # indexing_listener contains a `Select` object
        assert not isinstance(vector_index.indexing_listener, str)
        select = vector_index.indexing_listener.select

        # TODO: probably MongoDB queries should all have a common base class
        self.db.command(
            {
                "dropSearchIndex": select.table_or_collection.identifier,
                "name": vector_index.identifier,
            }
        )

    def disconnect(self):
        """
        Disconnect the client
        """

        # TODO: implement me



import abc
import asyncio
import dataclasses as dc
import inspect
import typing
from functools import reduce, wraps
from logging import WARNING, getLogger
from typing import Any, Callable, List, Optional, Union

from aladindb import logging
from aladindb.components.component import Component
from aladindb.components.model import _Predictor
from aladindb.ext.utils import format_prompt

if typing.TYPE_CHECKING:
    from aladindb.base.datalayer import Datalayer

# Disable httpx info level logging
getLogger("httpx").setLevel(WARNING)


def ensure_initialized(func):
    """Decorator to ensure that the model is initialized before calling the function"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if not hasattr(self, "_is_initialized") or not self._is_initialized:
            model_message = f"{self.__class__.__name__} : {self.identifier}"
            logging.info(f"Initializing {model_message}")
            self.init()
            self._is_initialized = True
            logging.info(f"Initialized  {model_message} successfully")
        return func(self, *args, **kwargs)

    return wrapper


@dc.dataclass
class _BaseLLM(Component, _Predictor, metaclass=abc.ABCMeta):
    """
    :param prompt_template: The template to use for the prompt.
    :param prompt_func: The function to use for the prompt.
    :param max_batch_size: The maximum batch size to use for batch generation.
    :param inference_kwargs: Parameters used during inference.
    """

    prompt_template: str = "{input}"
    prompt_func: Optional[Callable] = dc.field(default=None)
    max_batch_size: Optional[int] = 64
    inference_kwargs: dict = dc.field(default_factory=dict)

    def __post_init__(self):
        super().__post_init__()
        self.takes_context = True
        self.identifier = self.identifier.replace("/", "-")
        assert "{input}" in self.prompt_template, "Template must contain {input}"

    def to_call(self, X, *args, **kwargs):
        raise NotImplementedError

    def post_create(self, db: "Datalayer") -> None:
        # TODO: Do not make sense to add this logic here,
        # Need a auto DataType to handle this
        from aladindb.backends.ibis.data_backend import IbisDataBackend
        from aladindb.backends.ibis.field_types import dtype

        if isinstance(db.databackend, IbisDataBackend) and self.encoder is None:
            self.encoder = dtype("str")

        # since then the `.add` clause is not necessary
        output_component = db.databackend.create_model_table_or_collection(
            self  # type: ignore[arg-type]
        )

        if output_component is not None:
            db.add(output_component)

    @abc.abstractmethod
    def init(self):
        ...

    @abc.abstractmethod
    def _generate(self, prompt: str, **kwargs: Any) -> str:
        ...

    def _batch_generate(self, prompts: List[str], **kwargs: Any) -> List[str]:
        """
        Base method to batch generate text from a list of prompts.
        If the model can run batch generation efficiently, pls override this method.
        """
        return [self._generate(prompt, **kwargs) for prompt in prompts]

    @ensure_initialized
    def _predict(self, X: Union[str, List[str]], one: bool = False, **kwargs: Any):
        if isinstance(X, str):
            x = self.format_prompt(X, **kwargs)
            return self._generate(x, **kwargs)
        else:
            xs = [self.format_prompt(x, **kwargs) for x in X]
            return self._batch_generate(xs, **kwargs)

    def format_prompt(self, x, **kwargs):
        """
        Format a prompt with the given prompt function or template.
        :param x: The input to format.
        :param kwargs: Any additional arguments to pass to the prompt function.
        """
        if self.prompt_func is not None:
            return self.prompt_func(x, **self.get_kwargs(self.prompt_func, kwargs))

        return format_prompt(x, self.prompt_template, kwargs.get("context", None))

    def get_kwargs(self, func, *kwargs_list):
        """
        Get kwargs and object attributes that are in the function signature
        :param func (Callable): function to get kwargs for
        :param kwargs (list of dict): kwargs to filter
        """

        total_kwargs = reduce(lambda x, y: {**y, **x}, [self.dict(), *kwargs_list])
        sig = inspect.signature(func)
        new_kwargs = {}
        for k, v in total_kwargs.items():
            if k in sig.parameters:
                new_kwargs[k] = v
        return new_kwargs


@dc.dataclass
class BaseLLMAPI(_BaseLLM):
    """
    :param api_url: The URL for the API.
    {parent_doc}
    """

    __doc__ = __doc__.format(parent_doc=_BaseLLM.__doc__)

    api_url: str = dc.field(default="")

    def __post_init__(self):
        super().__post_init__()
        assert self.api_url, "api_url can not be empty"

    def init(self):
        pass


@dc.dataclass
class BaseOpenAI(_BaseLLM):
    """
    :param openai_api_base: The base URL for the OpenAI API.
    :param openai_api_key: The API key to use for the OpenAI API.
    :param model_name: The name of the model to use.
    :param chat: Whether to use the chat API or the completion API. Defaults to False.
    :param system_prompt: The prompt to use for the system.
    :param user_role: The role to use for the user.
    :param system_role: The role to use for the system.
    {parent_doc}
    """

    __doc__ = __doc__.format(parent_doc=_BaseLLM.__doc__)

    identifier: str = dc.field(default="")
    openai_api_base: Optional[str] = None
    openai_api_key: Optional[str] = None
    model_name: str = "gpt-3.5-turbo"
    chat: bool = True
    system_prompt: Optional[str] = None
    user_role: str = "user"
    system_role: str = "system"

    def __post_init__(self):
        self.identifier = self.identifier or self.model_name
        super().__post_init__()

    def init(self):
        try:
            from openai import AsyncClient, OpenAI
        except ImportError:
            raise Exception("You must install openai with command 'pip install openai'")

        params = {
            "api_key": self.openai_api_key,
            "base_url": self.openai_api_base,
        }

        self.client = OpenAI(**params)
        self.aclient = AsyncClient(**params)
        model_list = self.client.models.list()
        model_set = sorted({model.id for model in model_list.data})
        assert (
            self.model_name in model_set
        ), f"model_name {self.model_name} is not in model_set {model_set}"

    def _generate(self, prompt: str, **kwargs: Any) -> str:
        if self.chat:
            return self._chat_generate(prompt, **kwargs)
        else:
            return self._prompt_generate(prompt, **kwargs)

    def _batch_generate(self, prompts: List[str], **kwargs: Any) -> List[str]:
        """
        Use asyncio to batch generate text from a list of prompts.
        """
        return asyncio.run(self._async_batch_generate(prompts, **kwargs))

    def _prompt_generate(self, prompt: str, **kwargs: Any) -> str:
        """
        Generate a completion for a given prompt with prompt format.
        """
        completion = self.client.completions.create(
            model=self.model_name,
            prompt=prompt,
            **self.get_kwargs(
                self.client.completions.create, kwargs, self.inference_kwargs
            ),
        )
        return completion.choices[0].text

    def _chat_generate(self, content: str, **kwargs: Any) -> str:
        """
        Generate a completion for a given prompt with chat format.
        :param prompt: The prompt to generate a completion for.
        :param kwargs: Any additional arguments to pass to the API.
        """
        messages = kwargs.get("messages", [])

        if self.system_prompt:
            messages = [
                {"role": self.system_role, "content": self.system_prompt}
            ] + messages

        messages.append({"role": self.user_role, "content": content})
        completion = self.client.chat.completions.create(
            messages=messages,
            model=self.model_name,
            **self.get_kwargs(
                self.client.chat.completions.create, kwargs, self.inference_kwargs
            ),
        )
        return completion.choices[0].message.content

    async def _async_generate(self, semaphore, prompt: str, **kwargs) -> str:
        async with semaphore:
            if self.chat:
                return await self._async_chat_generate(prompt, **kwargs)
            else:
                return await self._async_prompt_generate(prompt, **kwargs)

    async def _async_prompt_generate(self, prompt: str, **kwargs) -> str:
        """
        Async method to generate a completion for a given prompt with prompt format.
        """
        completion = await self.aclient.completions.create(
            model=self.model_name,
            prompt=prompt,
            **self.get_kwargs(
                self.aclient.completions.create, kwargs, self.inference_kwargs
            ),
        )
        return completion.choices[0].text

    async def _async_chat_generate(self, content: str, **kwargs) -> str:
        """
        Async method to generate a completion for a given prompt with chat format.
        """
        messages = kwargs.get("messages", [])

        if self.system_prompt:
            messages = [
                {"role": self.system_role, "content": self.system_prompt}
            ] + messages

        messages.append({"role": self.user_role, "content": content})
        completion = await self.aclient.chat.completions.create(
            messages=messages,
            model=self.model_name,
            **self.get_kwargs(
                self.aclient.chat.completions.create, kwargs, self.inference_kwargs
            ),
        )
        return completion.choices[0].message.content

    async def _async_batch_generate(self, prompts: List[str], **kwargs) -> List[str]:
        """
        Async method to concurrently process multiple prompts.
        """
        semaphore = asyncio.Semaphore(self.max_batch_size or len(prompts))
        tasks = [
            self._async_generate(semaphore, prompt, **kwargs) for prompt in prompts
        ]
        return await asyncio.gather(*tasks)


@dc.dataclass
class BaseLLMModel(_BaseLLM):
    """
    :param model_name: The name of the model to use.
    :param on_ray: Whether to run the model on Ray.
    :param ray_config: The Ray config to use.
    :param ray_addredd: The address of the ray cluster.
    {parent_doc}
    """

    __doc__ = __doc__.format(parent_doc=_BaseLLM.__doc__)

    identifier: str = dc.field(default="")
    model_name: str = dc.field(default="")
    on_ray: bool = False
    ray_address: Optional[str] = None
    ray_config: dict = dc.field(default_factory=dict)

    def __post_init__(self):
        self.identifier = self.identifier or self.model_name
        assert self.model_name, "model_name can not be empty"
        self._is_initialized = False
        super().__post_init__()
